#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from __future__ import annotations

import importlib
import inspect
import logging
from importlib._bootstrap import ModuleSpec
from importlib.util import find_spec
from pathlib import Path
from typing import Dict, List, Optional

from health_ml.lightning_container import LightningContainer


class ModelConfigLoader:
    """
    Helper class to manage model config loading.
    """

    def __init__(self) -> None:
        pass

    def default_module_spec(self) -> ModuleSpec:
        from health_ml import configs  # type: ignore

        default_module = configs.__name__
        return find_spec(default_module)

    def find_module_search_specs(self, model_name: str) -> ModuleSpec:
        """
        Given model name (either only the class name or fully qualified), return the ModuleSpec that should be used for
        loading. If the model name is only the class name, the function will return the result of calling
        default_module_spec. Otherwise, this will return the module of the (fully qualified) model name.
        """
        model_namespace_parts = model_name.split(".")
        if len(model_namespace_parts) == 1:
            # config must be in the default path, nothing to be done
            return self.default_module_spec()

        module_name = ".".join(model_namespace_parts[:-1])
        logging.debug(f"Getting specification for module {module_name}")
        try:
            custom_spec: Optional[ModuleSpec] = find_spec(module_name)
        except Exception:
            custom_spec = None
        if custom_spec is None:
            raise ValueError(f"Module {module_name} was not found.")
        return custom_spec

    def _get_model_config(self, module_spec: ModuleSpec, model_name: str) -> Optional[LightningContainer]:
        """
        Given a module specification check to see if it has a class property with
        the <model_name> provided, and instantiate that config class with the
        provided <config_overrides>. Otherwise, return None.

        :param module_spec:
        :return: Instantiated model config if it was found.
        """
        # noinspection PyBroadException
        try:
            logging.debug(f"Importing {module_spec.name}")
            target_module = importlib.import_module(module_spec.name)
            # The "if" clause checks that obj is a class, of the desired name, that is
            # defined in this module rather than being imported into it (and hence potentially
            # being found twice).
            _class = next(
                obj
                for name, obj in inspect.getmembers(target_module)
                if inspect.isclass(obj) and name == model_name and inspect.getmodule(obj) == target_module
            )
            logging.info(f"Found class {_class} in file {module_spec.origin}")
        # ignore the exception which will occur if the provided module cannot be loaded
        # or the loaded module does not have the required class as a member
        except Exception as e:
            exception_text = str(e)
            if exception_text != "":
                logging.warning(f"Error when trying to import module {module_spec.name}: {exception_text}")
            return None
        model_config = _class()
        return model_config

    def _search_recursively_and_store(self, module_spec: ModuleSpec, model_name: str) -> Dict[str, LightningContainer]:
        """
        Given a root namespace eg: A.B.C searches recursively in all child namespaces
        for class property with the <model_name> provided. If found, this is
        instantiated with the provided overrides, and added to the configs dictionary.

        :param module_search_spec:
        """
        configs: Dict[str, LightningContainer] = {}
        root_namespace = module_spec.name
        namespaces_to_search: List[str] = []
        if module_spec.submodule_search_locations:
            logging.debug(
                f"Searching through {len(module_spec.submodule_search_locations)} folders that match namespace "
                f"{module_spec.name}: {module_spec.submodule_search_locations}"
            )
            for root in module_spec.submodule_search_locations:
                # List all python files in all the dirs under root, except for private dirs (prefixed with .)
                all_py_files = [x for x in Path(root).rglob("*.py") if ".." not in str(x)]
                for f in all_py_files:
                    if f.is_file() and "__pycache__" not in str(f) and f.name != "setup.py":
                        sub_namespace = path_to_namespace(f, root=root)
                        namespaces_to_search.append(root_namespace + "." + sub_namespace)
        elif module_spec.origin:
            # The module search spec already points to a python file: Search only that.
            namespaces_to_search.append(module_spec.name)
        else:
            raise ValueError(f"Unable to process module spec: {module_spec}")

        for n in namespaces_to_search:  # type: ignore
            _module_spec = None
            # noinspection PyBroadException
            try:
                _module_spec = find_spec(n)  # type: ignore
            except Exception:
                continue

            if _module_spec:
                config = self._get_model_config(_module_spec, model_name=model_name)
                if config:
                    configs[n] = config  # type: ignore
        return configs

    def create_model_config_from_name(self, model_name: str) -> LightningContainer:
        """
        Returns a model configuration for a model of the given name.

        :param model_name: Class name (for example, "HelloWorld") if the model config is in the default search
        namespace, or fully qualified name of the model, like mymodule.configs.MyConfig)
        """
        if not model_name:
            raise ValueError("Unable to load a model configuration because the model name is missing.")

        logging.info(f"Trying to locate model {model_name}")

        name_parts = model_name.split(".")
        class_name = name_parts[-1]
        module_spec = self.find_module_search_specs(model_name)
        configs = self._search_recursively_and_store(module_spec=module_spec, model_name=class_name)
        if len(configs) == 0:
            raise ValueError(f"Model '{model_name}' was not found in search namespace {module_spec.name}")
        elif len(configs) > 1:
            raise ValueError(
                f"Multiple instances of model '{model_name}' were found in namespaces: {[*configs.keys()]}"
            )
        else:
            return list(configs.values())[0]


def path_to_namespace(path: Path, root: Path) -> str:
    """
    Given a path (in form R/A/B/C) and a root directory R, create a namespace string A.B.C.
    The path must be located under the root directory.

    :param path: Path to convert to namespace
    :param root: Path prefix to remove from namespace.
    :return: A Python namespace string
    """
    return ".".join([Path(x).stem for x in path.relative_to(root).parts])



#  -------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  -------------------------------------------------------------------------------------------

import logging
import tempfile
from azureml.core import Run
from pathlib import Path
from typing import Optional

from health_azure import RUN_CONTEXT
from health_azure.utils import is_global_rank_zero, is_running_in_azure_ml
from health_ml.utils.common_utils import DEFAULT_AML_UPLOAD_DIR
from health_ml.lightning_container import LightningContainer
from health_ml.utils.checkpoint_utils import (
    download_highest_epoch_checkpoint,
    find_recovery_checkpoint_on_disk_or_cloud,
)


class CheckpointHandler:
    """
    This class handles which checkpoints are used to initialize the model during train or test time
    """

    def __init__(self, container: LightningContainer, project_root: Path, run_context: Optional[Run] = None):
        self.container = container
        self.project_root = project_root
        self.run_context = run_context
        self.trained_weights_path: Optional[Path] = None
        self.has_continued_training = False

    def download_recovery_checkpoints_or_weights(self) -> None:
        """
        Download checkpoints from a run recovery object or from a given checkpoint. Set the checkpoints path based on
        the checkpoint_url, local_checkpoint or checkpoint from an azureml run id.
        This is called at the start of training.
        """
        if self.container.src_checkpoint:
            self.trained_weights_path = self.container.src_checkpoint.get_or_download_checkpoint(
                download_dir=self.container.checkpoint_folder
            )
            self.container.trained_weights_path = self.trained_weights_path

    def additional_training_done(self) -> None:
        """
        Lets the handler know that training was done in this run.
        """
        self.has_continued_training = True

    def get_recovery_or_checkpoint_path_train(self) -> Optional[Path]:
        """
        Decides the checkpoint path to use for the current training run. Looks for the latest checkpoint in the
        checkpoint folder. If run_recovery is provided, the checkpoints will have been downloaded to this folder
        prior to calling this function. Else, if the run gets pre-empted and automatically restarted in AML,
        the latest checkpoint will be present in this folder too.

        :return: Constructed checkpoint path to recover from.
        """
        if is_global_rank_zero():
            checkpoints = list(self.container.checkpoint_folder.rglob("*"))
            logging.info(f"There are {len(checkpoints)} checkpoints in the checkpoint folder:")
            for f in checkpoints:
                logging.info(f.relative_to(self.container.checkpoint_folder))
        return find_recovery_checkpoint_on_disk_or_cloud(self.container.checkpoint_folder)

    def get_checkpoint_to_test(self) -> Path:
        """
        Find the model checkpoint that should be used for inference.

        If the model has been doing training epochs, get the best checkpoint as defined by the container.
        It is possible that the best checkpoint is not available on disk because the job got preempted. In those
        cases, try to find the inference checkpoint by going through all inference checkpoints stored in the AzureML
        run, downloading them and finding the one that has the highest epoch number (that must be the most recent
        among the possibly multiple retry results).

        If the model has not been doing training, but is set up to use a pre-trained
        set of weights in `trained_weights_path`, return that.
        """
        if self.has_continued_training:
            # If model was trained, look for the best checkpoint
            checkpoint_from_current_run = self.container.get_checkpoint_to_test()
            if checkpoint_from_current_run.is_file():
                logging.info(f"Using checkpoint from current run: {checkpoint_from_current_run}")
                return checkpoint_from_current_run
            logging.warning(f"No inference checkpoint found from the current run: {checkpoint_from_current_run}")
            logging.info("Trying to find an inference checkpoint in AzureML.")
            downloaded_checkpoint = self.download_inference_checkpoint()
            if downloaded_checkpoint is not None:
                logging.info(f"Using a checkpoint found in the AzureML run: {downloaded_checkpoint}")
                return downloaded_checkpoint
            raise FileNotFoundError("No inference checkpoint file found locally nor in AzureML.")
        elif self.trained_weights_path:
            # Model was not trained, check if there is a local weight path.
            logging.info(f"Using pre-trained weights from {self.trained_weights_path}")
            return self.trained_weights_path
        raise ValueError("Unable to determine which checkpoint should be used for testing.")

    def get_relative_inference_checkpoint_path(self) -> Path:
        """Returns the path of the model's inference checkpoint, relative to the container's output folder.

        This will be the path where the inference checkpoint can be found in the AzureML run output (except
        the `outputs` prefix that has to be added at the start)."""
        expected_checkpoint_path = self.container.get_checkpoint_to_test()
        try:
            return expected_checkpoint_path.relative_to(self.container.outputs_folder)
        except ValueError:
            raise ValueError(
                "Inference checkpoint path should be relative to the container's output folder. "
                f"Checkpoint path: {expected_checkpoint_path}, "
                f"output folder: {self.container.outputs_folder}"
            )

    def download_inference_checkpoint(self, download_folder: Optional[Path] = None) -> Optional[Path]:
        """
        For low-priority preemption that occured after training, try to download the inference checkpoint if that
        is available in the AzureML run from a previous incarnation of the job. Downloads go into the `download_folder`.

        The method returns None if no checkpoint was found, or if the
        current code is executing outside of AzureML and hence can't access previous inference checkpoints in AzureML.

        :param download_folder: The folder where the checkpoints should be downloaded to. If not provided, use a
            temp folder.
        :return: The path to a downloaded inference checkpoint, or None if no checkpoint was available.
        """
        # This logic will only trigger in AzureML. Download should only happen once per node.
        if is_running_in_azure_ml() and is_global_rank_zero():
            download_folder = download_folder or Path(tempfile.mkdtemp())
            inference_checkpoint_azureml_path = (
                Path(DEFAULT_AML_UPLOAD_DIR) / self.get_relative_inference_checkpoint_path()
            ).as_posix()
            highest_epoch_checkpoint = download_highest_epoch_checkpoint(
                run=RUN_CONTEXT, checkpoint_suffix=inference_checkpoint_azureml_path, output_folder=download_folder
            )
            if highest_epoch_checkpoint is None:
                logging.info("No inference checkpoint was found in the AzureML run.")
                return None
            logging.info("An inference checkpoint was found in the AzureML run.")
            return highest_epoch_checkpoint
        return None



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

from health_ml.utils.logging import AzureMLLogger, AzureMLProgressBar, log_learning_rate, log_on_epoch
from health_ml.utils.diagnostics import BatchTimeCallback
from health_ml.utils.common_utils import set_model_to_eval_mode
from health_ml.utils.package_setup import health_ml_package_setup


__all__ = [
    "AzureMLLogger",
    "AzureMLProgressBar",
    "BatchTimeCallback",
    "log_learning_rate",
    "log_on_epoch",
    "set_model_to_eval_mode",
    "health_ml_package_setup",
]



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import logging
import os
from pathlib import Path
from typing import Any, List, Optional, Tuple, TypeVar

from azureml.core import Run
from pytorch_lightning import Callback, Trainer
from pytorch_lightning.callbacks import GPUStatsMonitor, ModelCheckpoint
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.strategies import DDPStrategy
from pytorch_lightning.profiler import BaseProfiler, SimpleProfiler, AdvancedProfiler, PyTorchProfiler

from health_azure.utils import RUN_CONTEXT, is_running_in_azure_ml

from health_ml.lightning_container import LightningContainer
from health_ml.utils import AzureMLProgressBar
from health_ml.utils.common_utils import AUTOSAVE_CHECKPOINT_FILE_NAME, EXPERIMENT_SUMMARY_FILE
from health_ml.utils.diagnostics import TrainingDiagnoticsCallback
from health_ml.utils.lightning_loggers import StoringLogger, HimlMLFlowLogger


T = TypeVar('T')


def write_experiment_summary_file(config: Any, outputs_folder: Path) -> None:
    """
    Writes the given config to disk in plain text in the default output folder.
    """
    output = str(config)
    outputs_folder.mkdir(exist_ok=True, parents=True)
    dst = outputs_folder / EXPERIMENT_SUMMARY_FILE
    dst.write_text(output)
    logging.info(output)


def get_pl_profiler(pl_profiler: Optional[str], outputs_folder: Path) -> Optional[BaseProfiler]:
    if pl_profiler:
        pl_profilers = {"simple": SimpleProfiler, "advanced": AdvancedProfiler, "pytorch": PyTorchProfiler}
        if pl_profiler not in pl_profilers:
            raise ValueError(
                "Unsupported profiler. Please choose one of the following options: simple, advanced, "
                "pytorch. You can refer to https://pytorch-lightning.readthedocs.io/en/stable/advanced/"
                "profiler.html to learn more about each profiler. You can specify a custom profiler by "
                "overriding the default behavior of get_trainer_arguments() in your lightning container. "
                "You can find an example here https://github.com/microsoft/hi-ml/tree/main/docs/source/"
                "debugging.md#L145"
            )
        profiler = pl_profilers[pl_profiler](dirpath=outputs_folder / "profiler")
        return profiler
    else:
        return None


def create_lightning_trainer(
    container: LightningContainer,
    resume_from_checkpoint: Optional[Path] = None,
    num_nodes: int = 1,
    multiple_trainloader_mode: str = "max_size_cycle",
    azureml_run_for_logging: Optional[Run] = None,
    mlflow_run_for_logging: Optional[str] = None,
) -> Tuple[Trainer, StoringLogger]:
    """
    Creates a Pytorch Lightning Trainer object for the given model configuration. It creates checkpoint handlers
    and loggers. That includes a diagnostic logger for use in unit tests, that is also returned as the second
    return value.

    :param container: The container with model and data.
    :param resume_from_checkpoint: If provided, training resumes from this checkpoint point.
    :param num_nodes: The number of nodes to use in distributed training.
    :param azureml_run_for_logging: An optional AzureML Run object to which all metrics should be logged. Use this
        argument to log to AzureML when the training is happening outside of AzureML. If `azureml_run_for_logging` is
        None and the present code is running in AzureML, the current run is used.
    :return: A tuple [Trainer object, diagnostic logger]
    """
    logging.debug(f"resume_from_checkpoint: {resume_from_checkpoint}")
    num_gpus = container.num_gpus_per_node()
    effective_num_gpus = num_gpus * num_nodes
    strategy = None
    if effective_num_gpus == 0:
        accelerator = "cpu"
        devices = 1
        message = "CPU"
    else:
        accelerator = "gpu"
        devices = num_gpus
        message = f"{devices} GPU"
        if effective_num_gpus > 1:
            # Accelerator should be "ddp" when running large models in AzureML (when using DDP_spawn, we get out of
            # GPU memory).
            # Initialize the DDP plugin. The default for pl_find_unused_parameters is False. If True, the plugin
            # prints out lengthy warnings about the performance impact of find_unused_parameters.
            strategy = DDPStrategy(
                find_unused_parameters=container.pl_find_unused_parameters,
                static_graph=container.pl_static_graph,
            )
            message += "s per node with DDP"
    logging.info(f"Using {message}")
    tensorboard_logger = TensorBoardLogger(save_dir=str(container.logs_folder), name="Lightning", version="")
    loggers: List[Any] = [tensorboard_logger]

    if is_running_in_azure_ml():
        mlflow_run_id = os.environ.get("MLFLOW_RUN_ID", None)
        logging.info(f"Logging to MLFlow run with id: {mlflow_run_id}")
        mlflow_logger = HimlMLFlowLogger(run_id=mlflow_run_id)
        loggers.append(mlflow_logger)
    else:
        mlflow_run_dir = container.outputs_folder / "mlruns"
        try:
            mlflow_run_dir.mkdir(exist_ok=True)
            mlflow_tracking_uri = "file:" + str(mlflow_run_dir)
            mlflow_logger = HimlMLFlowLogger(run_id=mlflow_run_for_logging, tracking_uri=mlflow_tracking_uri)
            loggers.append(mlflow_logger)
            logging.info(
                f"Logging to MLFlow run with id: {mlflow_run_for_logging}. Local MLFlow logs are stored in "
                f"{mlflow_tracking_uri}"
            )
        except FileNotFoundError as e:
            logging.warning(f"Unable to initialise MLFlowLogger due to error: {e}")

    storing_logger = StoringLogger()
    loggers.append(storing_logger)
    # Use 32bit precision when running on CPU. Otherwise, make it depend on use_mixed_precision flag.
    precision = 32 if num_gpus == 0 else 16 if container.use_mixed_precision else 32
    # The next two flags control the settings in torch.backends.cudnn.deterministic and torch.backends.cudnn.benchmark
    # https://pytorch.org/docs/stable/notes/randomness.html
    # Note that switching to deterministic models can have large performance downside.
    if container.pl_deterministic:
        deterministic = True
        benchmark = False
    else:
        deterministic = False
        benchmark = True

    # The last checkpoint is considered the "best" checkpoint. For large segmentation
    # models, this still appears to be the best way of choosing them because validation loss on the relatively small
    # training patches is not stable enough. Going by the validation loss somehow works for the Prostate model, but
    # not for the HeadAndNeck model.
    # Note that "last" is somehow a misnomer, it should rather be "latest". There is a "last" checkpoint written in
    # every epoch. We could use that for recovery too, but it could happen that the job gets preempted right during
    # writing that file, and we would end up with an invalid file.
    last_checkpoint_callback = ModelCheckpoint(dirpath=str(container.checkpoint_folder), save_last=True, save_top_k=0)
    recovery_checkpoint_callback = ModelCheckpoint(
        dirpath=str(container.checkpoint_folder),
        filename=AUTOSAVE_CHECKPOINT_FILE_NAME,
        every_n_epochs=container.autosave_every_n_val_epochs,
        save_last=False,
    )
    callbacks: List[Callback] = [
        last_checkpoint_callback,
        recovery_checkpoint_callback,
    ]
    if container.monitor_loading:
        # TODO antonsc: Remove after fixing the callback.
        raise NotImplementedError("Monitoring batch loading times has been temporarily disabled.")
        # callbacks.append(BatchTimeCallback())
    if container.monitor_training:
        callbacks.append(TrainingDiagnoticsCallback())
    if num_gpus > 0 and container.monitor_gpu:
        logging.info("Adding monitoring for GPU utilization")
        callbacks.append(GPUStatsMonitor(intra_step_time=True, inter_step_time=True))
    # Add the additional callbacks that were specified in get_trainer_arguments for LightningContainers
    additional_args = container.get_trainer_arguments()
    # Callbacks can be specified via the "callbacks" argument (the legacy behaviour) or the new get_callbacks method
    if "callbacks" in additional_args:
        more_callbacks = additional_args.pop("callbacks")
        if isinstance(more_callbacks, list):
            callbacks.extend(more_callbacks)  # type: ignore
        else:
            callbacks.append(more_callbacks)  # type: ignore
    callbacks.extend(container.get_callbacks())
    # Set profiler: if --pl_profiler=profiler is specified, it overrides any custom profiler defined in the container
    custom_profiler = additional_args.pop("profiler", None)
    profiler = get_pl_profiler(container.pl_profiler, container.outputs_folder)
    profiler = profiler if profiler else custom_profiler
    is_azureml_run = is_running_in_azure_ml(RUN_CONTEXT)
    progress_bar_refresh_rate = container.pl_progress_bar_refresh_rate
    if progress_bar_refresh_rate is None:
        progress_bar_refresh_rate = 50
        logging.info(
            f"The progress bar refresh rate is not set. Using a default of {progress_bar_refresh_rate}. "
            f"To change, modify the pl_progress_bar_refresh_rate field of the container."
        )
    if is_azureml_run:
        callbacks.append(
            AzureMLProgressBar(
                refresh_rate=progress_bar_refresh_rate, write_to_logging_info=True, print_timestamp=False
            )
        )
    else:
        # Use a local import here to be able to support older PL versions
        from pytorch_lightning.callbacks import TQDMProgressBar

        callbacks.append(TQDMProgressBar(refresh_rate=progress_bar_refresh_rate))
    # Read out additional model-specific args here.
    # We probably want to keep essential ones like numgpu and logging.
    trainer = Trainer(
        default_root_dir=str(container.outputs_folder),
        deterministic=deterministic,
        benchmark=benchmark,
        accelerator=accelerator,
        strategy=strategy,
        max_epochs=container.max_epochs,
        # All of the following limit_batches  arguments can be integers or floats.
        # If integers, it is the number of batches.
        # If float, it's the fraction of batches. We default to 1.0 (processing all batches).
        limit_train_batches=container.pl_limit_train_batches or 1.0,
        limit_val_batches=container.pl_limit_val_batches or 1.0,
        limit_test_batches=container.pl_limit_test_batches or 1.0,
        fast_dev_run=container.pl_fast_dev_run,  # type: ignore
        num_sanity_val_steps=container.pl_num_sanity_val_steps,
        log_every_n_steps=container.pl_log_every_n_steps,
        # check_val_every_n_epoch=container.pl_check_val_every_n_epoch,
        callbacks=callbacks,
        logger=loggers,
        num_nodes=num_nodes,
        devices=devices,
        precision=precision,
        sync_batchnorm=container.pl_sync_batchnorm,
        detect_anomaly=container.detect_anomaly,
        profiler=profiler,
        resume_from_checkpoint=str(resume_from_checkpoint) if resume_from_checkpoint else None,
        multiple_trainloader_mode=multiple_trainloader_mode,
        accumulate_grad_batches=container.pl_accumulate_grad_batches,
        replace_sampler_ddp=container.pl_replace_sampler_ddp,
        **additional_args,
    )
    return trainer, storing_logger



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from __future__ import annotations

import logging
import os
import param
from enum import Enum, unique
from param import Parameterized
from pathlib import Path
from typing import Any, Dict, List, Optional

from azureml.train.hyperdrive import HyperDriveConfig

from health_azure import create_crossval_hyperdrive_config
from health_azure.himl import (
    create_grid_hyperdrive_config,
    create_crossval_hyperparam_args_v2,
    create_grid_hyperparam_args_v2,
)
from health_azure.amulet import (
    ENV_AMLT_PROJECT_NAME,
    ENV_AMLT_INPUT_OUTPUT,
    ENV_AMLT_SNAPSHOT_DIR,
    ENV_AMLT_AZ_BATCHAI_DIR,
    is_amulet_job,
    get_amulet_aml_working_dir,
)
from health_azure.utils import RUN_CONTEXT, PathOrString, is_global_rank_zero, is_running_in_azure_ml
from health_ml.utils import fixed_paths
from health_ml.utils.checkpoint_utils import CheckpointParser
from health_ml.utils.common_utils import (
    CHECKPOINT_FOLDER,
    create_unique_timestamp_id,
    DEFAULT_AML_UPLOAD_DIR,
    DEFAULT_LOGS_DIR_NAME,
)
from health_ml.utils.type_annotations import IntOrFloat, TupleFloat2


@unique
class LRWarmUpType(Enum):
    """
    Supported LR warm up types for model training
    """

    NoWarmUp = "NoWarmUp"
    Linear = "Linear"


@unique
class LRSchedulerType(Enum):
    """
    Supported lr scheduler types for model training
    """

    Exponential = "Exponential"
    Step = "Step"
    Polynomial = "Polynomial"
    Cosine = "Cosine"
    MultiStep = "MultiStep"


@unique
class OptimizerType(Enum):
    """
    Supported optimizers for model training
    """

    Adam = "Adam"
    AMSGrad = "AMSGrad"
    SGD = "SGD"
    RMSprop = "RMSprop"


class ExperimentFolderHandler(Parameterized):
    """High level config to abstract the file system related settings for experiments"""

    outputs_folder: Path = param.ClassSelector(
        class_=Path, default=Path(), instantiate=False, doc="The folder where all training and test outputs should go."
    )
    logs_folder: Path = param.ClassSelector(
        class_=Path, default=Path(), instantiate=False, doc="The folder for all log files and Tensorboard event files"
    )
    project_root: Path = param.ClassSelector(
        class_=Path,
        default=Path(),
        instantiate=False,
        doc="The root folder for the codebase that triggers the training run.",
    )
    run_folder: Path = param.ClassSelector(
        class_=Path, default=Path(), instantiate=False, doc="The folder that contains outputs and the logs subfolder."
    )

    @staticmethod
    def create(
        project_root: Path, is_offline_run: bool, model_name: str, output_to: Optional[Path] = None
    ) -> ExperimentFolderHandler:
        """
        Creates a new object that holds output folder configurations. When running inside of AzureML, the output
        folders will be directly under the project root. If not running inside AzureML, a folder with a timestamp
        will be created for all outputs and logs.

        :param project_root: The root folder that contains the code that submitted the present training run.
        When running inside the hi-ml repository, it is the git repo root. When consuming hi-ml as a package,
        this should be the root of the source code that calls the package.
        :param is_offline_run: If true, this is a run outside AzureML. If False, it is inside AzureML.
        :param model_name: The name of the model that is trained. This is used to generate a run-specific output
        folder.
        :param output_to: If provided, the output folders will be created as a subfolder of this argument. If not
        given, the output folders will be created inside of the project root.
        """
        if not project_root.is_absolute():
            raise ValueError(f"The project root is required to be an absolute path, but got {project_root}")
        if is_offline_run or output_to:
            if output_to:
                logging.info(f"All results will be written to the specified output folder {output_to}")
                root = Path(output_to).resolve()
            else:
                logging.info("All results will be written to a subfolder of the project root folder.")
                root = project_root.resolve() / DEFAULT_AML_UPLOAD_DIR
            if is_global_rank_zero():
                timestamp = create_unique_timestamp_id()
                run_folder = root / f"{timestamp}_{model_name}"
            else:
                # Handle the case where there are multiple DDP threads on the same machine outside AML.
                # Each child process will be started with the current working directory set to be the output
                # folder of the rank 0 process. We want all other process to write to that same folder.
                run_folder = Path.cwd()
            outputs_folder = run_folder
            logs_folder = run_folder / DEFAULT_LOGS_DIR_NAME
        else:
            logging.info("Running inside AzureML.")
            logging.info("All results will be written to a subfolder of the project root folder.")
            if not is_amulet_job():
                run_folder = project_root
                outputs_folder = project_root / DEFAULT_AML_UPLOAD_DIR
                logs_folder = project_root / DEFAULT_LOGS_DIR_NAME
            else:
                # Job submitted via Amulet
                amlt_root_folder = Path(os.environ[ENV_AMLT_INPUT_OUTPUT])
                project_name = os.environ[ENV_AMLT_PROJECT_NAME]
                snapshot_dir = get_amulet_aml_working_dir()
                assert (
                    snapshot_dir
                ), f"Either {ENV_AMLT_SNAPSHOT_DIR} or {ENV_AMLT_AZ_BATCHAI_DIR} must exist in env vars"
                print(
                    f"Found the following environment variables set by Amulet: "
                    f"AZURE_ML_INPUT_OUTPUT: {amlt_root_folder}, AZUREML_ARM_PROJECT_NAME: {project_name}"
                )
                run_id = RUN_CONTEXT.id
                run_folder = amlt_root_folder / "projects" / project_name / "amlt-code" / run_id
                outputs_folder = snapshot_dir / DEFAULT_AML_UPLOAD_DIR
                logs_folder = snapshot_dir / DEFAULT_LOGS_DIR_NAME

        logging.info(f"Run outputs folder: {outputs_folder}")
        logging.info(f"Logs folder: {logs_folder}")
        logging.info(f"Run root directory: {run_folder}")
        return ExperimentFolderHandler(
            outputs_folder=outputs_folder, logs_folder=logs_folder, project_root=project_root, run_folder=run_folder
        )


class WorkflowParams(param.Parameterized):
    """
    This class contains all parameters that affect how the whole training and testing workflow is executed.
    """

    random_seed: int = param.Integer(42, doc="The seed to use for all random number generators.")
    src_checkpoint: CheckpointParser = param.ClassSelector(
        class_=CheckpointParser, default=None, instantiate=False, doc=CheckpointParser.DOC
    )
    crossval_count: int = param.Integer(
        default=1,
        bounds=(0, None),
        doc="The number of splits to use when doing cross-validation. Use 1 to disable cross-validation",
    )
    crossval_index: Optional[int] = param.Integer(
        default=None,
        bounds=(0, None),
        allow_None=True,
        doc="When doing cross validation, this is the index of the current "
        "split. Valid values: 0 .. (crossval_count -1). Default: None"
        "when launching a cross-validation run, this will be set by "
        "the hyperdrive config. To submit a single cross-validation fold,"
        "set this to the fold index along with --crossval_count > 1.",
    )
    hyperdrive: bool = param.Boolean(
        False,
        doc="If True, use the Hyperdrive configuration specified in the "
        "LightningContainer to run hyperparameter tuning. If False, just "
        "run a plain single training job. This cannot be combined with "
        "the flags --different_seeds or --crossval_count.",
    )
    regression_test_folder: Optional[Path] = param.ClassSelector(
        class_=Path,
        default=None,
        allow_None=True,
        doc="A path to a folder that contains a set of files. At the end of training and "
        "model evaluation, all files given in that folder must be present in the job's output "
        "folder, and their contents must match exactly. When running in AzureML, you need to "
        "ensure that this folder is part of the snapshot that gets uploaded. The path should "
        "be relative to the repository root directory.",
    )
    regression_test_csv_tolerance: float = param.Number(
        default=0.0,
        allow_None=False,
        doc="When comparing CSV files during regression tests, use this value as the maximum allowed "
        "relative difference of actual and expected results. Default: 0.0 (must match exactly)",
    )
    regression_metrics: str = param.String(default=None, doc="A list of names of metrics to compare")
    run_inference_only: bool = param.Boolean(
        False,
        doc="If True, run only inference and skip training after loading"
        "model weights from the specified checkpoint in "
        "`src_checkpoint` flag. Inference is run on both validation "
        "and test sets. If False, run training and inference.",
    )
    resume_training: bool = param.Boolean(False, doc="If True, resume training from the src_checkpoint.")
    tag: str = param.String(doc="A string that will be used as the display name of the run in AzureML.")
    experiment: str = param.String(
        default="",
        doc="The name of the AzureML experiment to use for this run. If not "
        "provided, the name of the model class will be used.",
    )
    log_from_vm: bool = param.Boolean(
        False,
        doc="If True, a training run outside AzureML will still log its "
        "metrics to AzureML. Both intermediate validation metrics and final test results"
        "will be recorded. You need to have an AzureML workspace config.json file "
        "and will be asked for interactive authentication.",
    )
    different_seeds: int = param.Integer(
        default=0,
        bounds=(0, None),
        doc="If > 0, run the same training job multiple times with different seeds. "
        "This uses AzureML hyperdrive to run multiple jobs in parallel, and hence "
        "cannot be used when running outside AzureML. "
        "This cannot be combined with the --hyperdrive or the --crossval_count flags.",
    )

    CROSSVAL_INDEX_ARG_NAME = "crossval_index"
    CROSSVAL_COUNT_ARG_NAME = "crossval_count"
    RANDOM_SEED_ARG_NAME = "random_seed"

    def validate(self) -> None:
        if self.crossval_count > 1 and self.crossval_index is not None:
            if not (0 <= self.crossval_index < self.crossval_count):
                raise ValueError(f"Attribute crossval_index out of bounds (crossval_count = {self.crossval_count})")

        if self.run_inference_only and not self.src_checkpoint:
            raise ValueError(f"Cannot run inference without a src_checkpoint. {CheckpointParser.INFO_MESSAGE}")
        if self.resume_training and not self.src_checkpoint:
            raise ValueError(f"Cannot resume training without a src_checkpoint. {CheckpointParser.INFO_MESSAGE}")

    @property
    def is_running_in_aml(self) -> bool:
        """
        Whether the current run is executing inside Azure ML

        :return: True if the run is executing inside Azure ML, or False if outside AzureML.
        """
        return is_running_in_azure_ml(RUN_CONTEXT)

    def get_effective_random_seed(self) -> int:
        """
        Returns the random seed set as part of this configuration. If the configuration corresponds
        to a cross validation split, then the cross validation fold index will be added to the
        set random seed in order to return the effective random seed.
        :return:
        """
        seed = self.random_seed
        if self.is_crossvalidation_child_run:
            # Offset the random seed based on the cross validation split index so each
            # fold has a different initial random state. Cross validation index 0 will have
            # a different seed from a non cross validation run.
            assert self.crossval_index is not None, "crossval_index must be set for cross validation child runs"
            seed += self.crossval_index + 1
        return seed

    @property
    def is_crossvalidation_parent_run(self) -> bool:
        """
        Returns True if the present parameters indicate that this is cross-validation parent run that should trigger
        a hyperdrive run.
        """
        return self.crossval_count > 1 and self.crossval_index is None

    @property
    def is_crossvalidation_child_run(self) -> bool:
        """
        Returns True if the present parameters indicate that this is a cross-validation child run.
        """
        return self.crossval_count > 1 and self.crossval_index is not None

    def get_crossval_hyperdrive_config(self) -> HyperDriveConfig:
        # For crossvalidation, the name of the metric to monitor does not matter because no early termination or such
        # is specified.
        return create_crossval_hyperdrive_config(
            num_splits=self.crossval_count,
            cross_val_index_arg_name=self.CROSSVAL_INDEX_ARG_NAME,
            metric_name="val/loss",
        )

    def get_different_seeds_hyperdrive_config(self) -> HyperDriveConfig:
        """Returns a configuration object for AzureML Hyperdrive that varies the random seed for each run."""
        return create_grid_hyperdrive_config(
            values=list(map(str, range(self.different_seeds))),
            argument_name=self.RANDOM_SEED_ARG_NAME,
            metric_name="val/loss",
        )

    def get_crossval_hyperparam_args_v2(self) -> Dict[str, Any]:
        """
        Wrapper function to create hyperparameter search arguments specifically for running cross validation
        with AML SDK v2

        :return: A dictionary of hyperparameter search arguments and values.
        """
        return create_crossval_hyperparam_args_v2(
            num_splits=self.crossval_count,
            cross_val_index_arg_name=self.CROSSVAL_INDEX_ARG_NAME,
            metric_name="val/loss",
        )

    def get_grid_hyperparam_args_v2(self) -> Dict[str, Any]:
        """
        Wrapper function to create hyperparameter search arguments specifically for running grid search
        with AML SDK v2

        :return: A dictionary of hyperparameter search arguments and values.
        """
        return create_grid_hyperparam_args_v2(
            values=list(map(str, range(self.different_seeds))),
            argument_name=self.RANDOM_SEED_ARG_NAME,
            metric_name="val/loss",
        )


class DatasetParams(param.Parameterized):
    datastore: str = param.String(default="", doc="Datastore to look for data in")
    azure_datasets: List[str] = param.List(
        default=[],
        class_=str,
        doc="If provided, the ID of one or more datasets to use when running in"
        " AzureML. This dataset must exist as a folder of the same name "
        "in the 'datasets' container in the datasets storage account. This "
        "dataset will be mounted and made available at the 'local_dataset' "
        "path when running in AzureML.",
    )
    local_datasets: List[Path] = param.List(
        default=[],
        class_=Path,
        doc="A list of one or more paths to the dataset to use, when training outside of Azure ML.",
    )
    dataset_mountpoints: List[Path] = param.List(
        default=[],
        class_=Path,
        doc="The path at which the AzureML dataset should be made "
        "available via mounting or downloading. This only affects "
        "jobs running in AzureML. If empty, use a random "
        "mount/download point.",
    )

    def validate(self) -> None:
        if (not self.azure_datasets) and (not self.local_datasets):
            raise ValueError("Either local_datasets or azure_datasets must be set.")

        if self.dataset_mountpoints and len(self.azure_datasets) != len(self.dataset_mountpoints):
            raise ValueError(
                f"Expected the number of azure datasets to equal the number of mountpoints, "
                f"got datasets [{','.join(self.azure_datasets)}] "
                f"and mountpoints [{','.join([str(m) for m in self.dataset_mountpoints])}]"
            )


class OutputParams(param.Parameterized):
    output_to: Optional[Path] = param.ClassSelector(
        class_=Path,
        default=None,
        doc="If provided, the run outputs will be written to the given "
        "folder. If not provided, outputs will go into a subfolder "
        "of the project root folder.",
    )
    file_system_config: ExperimentFolderHandler = param.ClassSelector(
        default=ExperimentFolderHandler(),
        class_=ExperimentFolderHandler,
        instantiate=False,
        doc="File system related configs",
    )
    _model_name: str = param.String(
        "", doc="The human readable name of the model (for example, Liver). This is usually set from the class name."
    )

    @property
    def model_name(self) -> str:
        """
        Gets the human readable name of the model (e.g., Liver). This is usually set from the class name.

        :return: A model name as a string.
        """
        return self._model_name

    def set_output_to(self, output_to: PathOrString) -> None:
        """
        Adjusts the file system settings in the present object such that all outputs are written to the given
        folder.

        :param output_to: The absolute path to a folder that should contain the outputs.
        """
        self.output_to = Path(output_to)
        self.create_filesystem(project_root=fixed_paths.repository_root_directory())

    def create_filesystem(self, project_root: Path) -> None:
        """
        Creates new file system settings (outputs folder, logs folder) based on the information stored in the
        present object. If any of the folders do not yet exist, they are created.

        :param project_root: The root folder for the codebase that triggers the training run.
        """
        self.file_system_config = ExperimentFolderHandler.create(
            project_root=project_root,
            model_name=self.model_name,
            is_offline_run=not is_running_in_azure_ml(RUN_CONTEXT),
            output_to=self.output_to,
        )

    @property
    def outputs_folder(self) -> Path:
        """Gets the full path in which the model outputs should be stored."""
        return self.file_system_config.outputs_folder

    @property
    def logs_folder(self) -> Path:
        """Gets the full path in which the model logs should be stored."""
        return self.file_system_config.logs_folder

    @property
    def checkpoint_folder(self) -> Path:
        """Gets the full path in which the model checkpoints should be stored during training."""
        return self.outputs_folder / CHECKPOINT_FOLDER


class OptimizerParams(param.Parameterized):
    l_rate: float = param.Number(1e-4, doc="The initial learning rate", bounds=(0, None))
    _min_l_rate: float = param.Number(
        0.0, doc="The minimum learning rate for the Polynomial and Cosine schedulers.", bounds=(0.0, None)
    )
    l_rate_scheduler: LRSchedulerType = param.ClassSelector(
        default=LRSchedulerType.Polynomial,
        class_=LRSchedulerType,
        instantiate=False,
        doc="Learning rate decay method (Cosine, Polynomial, Step, MultiStep or Exponential)",
    )
    l_rate_exponential_gamma: float = param.Number(
        0.9, doc="Controls the rate of decay for the Exponential LR scheduler."
    )
    l_rate_step_gamma: float = param.Number(0.1, doc="Controls the rate of decay for the Step LR scheduler.")
    l_rate_step_step_size: int = param.Integer(50, bounds=(0, None), doc="The step size for Step LR scheduler")
    l_rate_multi_step_gamma: float = param.Number(0.1, doc="Controls the rate of decay for the MultiStep LR scheduler.")
    l_rate_multi_step_milestones: Optional[List[int]] = param.List(
        None, bounds=(1, None), allow_None=True, class_=int, doc="The milestones for MultiStep decay."
    )
    l_rate_polynomial_gamma: float = param.Number(
        1e-4, doc="Controls the rate of decay for the Polynomial LR scheduler."
    )
    l_rate_warmup: LRWarmUpType = param.ClassSelector(
        default=LRWarmUpType.NoWarmUp,
        class_=LRWarmUpType,
        instantiate=False,
        doc="The type of learning rate warm up to use. Can be NoWarmUp (default) or Linear.",
    )
    l_rate_warmup_epochs: int = param.Integer(
        0,
        bounds=(0, None),
        doc="Number of warmup epochs (linear warmup) before the "
        "scheduler starts decaying the learning rate. "
        "For example, if you are using MultiStepLR with "
        "milestones [50, 100, 200] and warmup epochs = 100, warmup "
        "will last for 100 epochs and the first decay of LR "
        "will happen on epoch 150",
    )
    optimizer_type: OptimizerType = param.ClassSelector(
        default=OptimizerType.Adam, class_=OptimizerType, instantiate=False, doc="The optimizer_type to use"
    )
    opt_eps: float = param.Number(1e-4, doc="The epsilon parameter of RMSprop or Adam")
    rms_alpha: float = param.Number(0.9, doc="The alpha parameter of RMSprop")
    adam_betas: TupleFloat2 = param.NumericTuple(
        (0.9, 0.999), length=2, doc="The betas parameter of Adam, default is (0.9, 0.999)"
    )
    momentum: float = param.Number(0.6, doc="The momentum parameter of the optimizers")
    weight_decay: float = param.Number(1e-4, doc="The weight decay used to control L2 regularization")

    def validate(self) -> None:
        if len(self.adam_betas) < 2:
            raise ValueError(
                "The adam_betas parameter should be the coefficients used for computing running averages of "
                "gradient and its square"
            )

        if self.l_rate_scheduler == LRSchedulerType.MultiStep:
            if not self.l_rate_multi_step_milestones:
                raise ValueError("Must specify l_rate_multi_step_milestones to use LR scheduler MultiStep")
            if sorted(set(self.l_rate_multi_step_milestones)) != self.l_rate_multi_step_milestones:
                raise ValueError("l_rate_multi_step_milestones must be a strictly increasing list")
            if self.l_rate_multi_step_milestones[0] <= 0:
                raise ValueError("l_rate_multi_step_milestones cannot be negative or 0.")

    @property
    def min_l_rate(self) -> float:
        return self._min_l_rate

    @min_l_rate.setter
    def min_l_rate(self, value: float) -> None:
        if value > self.l_rate:
            raise ValueError("l_rate must be >= min_l_rate, found: {}, {}".format(self.l_rate, value))
        self._min_l_rate = value


class TrainerParams(param.Parameterized):
    max_epochs: int = param.Integer(100, bounds=(1, None), doc="Number of epochs to train.")
    autosave_every_n_val_epochs: int = param.Integer(
        1,
        bounds=(0, None),
        doc="Save epoch checkpoints every N validation epochs. "
        "If pl_check_val_every_n_epoch > 1, this means that "
        "checkpoints are saved every "
        "N * pl_check_val_every_n_epoch training epochs.",
    )
    detect_anomaly: bool = param.Boolean(
        False, doc="If true, test gradients for anomalies (NaN or Inf) during training."
    )
    use_mixed_precision: bool = param.Boolean(
        True,
        doc="If True, use float16 precision (Native Adaptive Mixed Precision) during training. "
        "If False, use float32.",
    )
    max_num_gpus: int = param.Integer(
        default=-1,
        doc="The maximum number of GPUS to use. If set to a value < 0, use"
        "all available GPUs. In distributed training, this is the "
        "maximum number of GPUs per node.",
    )
    max_num_gpus_inference: int = param.Integer(
        default=1,
        doc="The maximum number of GPUS to use for inference. Default is 1 for "
        "single device inference. This guarantees reproducibility of "
        "results without any duplication of data. However, if you use "
        "pl_replace_sampler_ddp=False, you can set this to a higher value "
        "to speed up inference. Use a negative value to use all available "
        "GPUs in the system.",
    )
    pl_progress_bar_refresh_rate: Optional[int] = param.Integer(
        default=None,
        doc="PyTorch Lightning trainer flag 'progress_bar_refresh_rate': How often to refresh "
        "progress bar (in steps). Value 0 disables progress bar. If None choose, automatically.",
    )
    pl_num_sanity_val_steps: int = param.Integer(
        default=0,
        doc="PyTorch Lightning trainer flag 'num_sanity_val_steps': Number of validation "
        "steps to run before training, to identify possible problems",
    )
    pl_deterministic: bool = param.Boolean(
        default=False,
        doc="Controls the PyTorch Lightning trainer flags 'deterministic' and 'benchmark'. If "
        "'pl_deterministic' is True, results are perfectly reproducible. If False, they are not, "
        "but you may see training speed increases.",
    )
    pl_find_unused_parameters: bool = param.Boolean(
        default=False,
        doc="Controls the PyTorch Lightning flag 'find_unused_parameters' for the DDP plugin. "
        "Setting it to True comes with a performance hit.",
    )
    pl_limit_train_batches: Optional[IntOrFloat] = param.Number(
        default=None,
        doc="PyTorch Lightning trainer flag 'limit_train_batches': Limit the training dataset to the "
        "given number of batches if integer, or proportion of training dataset if float.",
    )
    pl_limit_val_batches: Optional[IntOrFloat] = param.Number(
        default=None,
        doc="PyTorch Lightning trainer flag 'limit_val_batches': Limit the validation dataset to the "
        "given number of batches if integer, or proportion of validation dataset if float.",
    )
    pl_limit_test_batches: Optional[IntOrFloat] = param.Number(
        default=None,
        doc="PyTorch Lightning trainer flag 'limit_test_batches': Limit the test dataset to the "
        "given number of batches if integer, or proportion of test dataset if float.",
    )
    pl_fast_dev_run: Optional[int] = param.Integer(
        default=0,
        doc="PyTorch Lightning trainer flag 'fast_dev_run': Runs n if set to 'n' batch(es) of train, val"
        "and test. Default to 0 to use all train, val and test batches available. Setting "
        "pl_fast_dev_run to n > 0 overrides pl_limit_{train, val, test}_batches to the same value n."
        "Additionally, by setting this argument, ALL (tuner, checkpoint callbacks, early stopping "
        "callbacks, loggers and loggger callbacks) will be disabled and run for only a single epoch."
        "This must be used only for debbuging purposes.",
    )
    pl_profiler: Optional[str] = param.String(
        default=None,
        doc="The value to use for the 'profiler' argument for the Lightning trainer. "
        "Set to either 'simple', 'advanced', or 'pytorch'",
    )
    pl_sync_batchnorm: bool = param.Boolean(
        default=True,
        doc="PyTorch Lightning trainer flag 'sync_batchnorm': If True, "
        "synchronize batchnorm across all GPUs when running in ddp mode."
        "If False, batchnorm is not synchronized.",
    )
    pl_static_graph: bool = param.Boolean(
        default=False, doc="PyTorch DDP flag 'static_graph': If True, the trained graph is static. Default is False."
    )
    monitor_gpu: bool = param.Boolean(
        default=False,
        doc="If True, add the GPUStatsMonitor callback to the Lightning trainer object. "
        "This will write GPU utilization metrics every 50 batches by default.",
    )
    monitor_loading: bool = param.Boolean(
        default=False,
        doc="If True, add the BatchTimeCallback callback to the Lightning trainer "
        "object. This will monitor how long individual batches take to load.",
    )
    monitor_training: bool = param.Boolean(
        default=False,
        doc="If True, add the TrainingDiagnosisCallback to the Lightning trainer "
        "object. This will monitor when training, validation and test starts "
        "and ends and also intermediate steps.",
    )
    run_extra_val_epoch: bool = param.Boolean(
        default=False,
        doc="If True, run an additional validation epoch at the end of training "
        "to produce plots outputs on the validation set. This is to reduce "
        "any validation overheads during training time and produce "
        "additional time or memory consuming outputs only once after "
        "training is finished on the validation set.",
    )
    pl_accumulate_grad_batches: int = param.Integer(
        default=1,
        doc="The number of batches over which gradients are accumulated, before a parameter update is done.",
    )
    pl_log_every_n_steps: int = param.Integer(
        default=50,
        doc="PyTorch Lightning trainer flag 'log_every_n_steps': How often to log within steps. Default to 50.",
    )
    pl_replace_sampler_ddp: bool = param.Boolean(
        default=True,
        doc="PyTorch Lightning trainer flag 'replace_sampler_ddp' that "
        "sets the sampler for distributed training with shuffle=True during "
        "training and shuffle=False during validation. Default to True. Set to"
        "False to set your own sampler.",
    )

    def validate(self) -> None:
        if self.max_num_gpus_inference > 1 and self.pl_replace_sampler_ddp:
            logging.warning(
                "The 'pl_replace_sampler_ddp' flag is set to True, but the 'max_num_gpus_inference' > 1. "
                "This might bias the inference results due to duplicate samples. Consider setting "
                "--pl_replace_sampler_ddp=False if you want to use multiple GPUs for inference."
                "pytorch_lightning.overrides.distributed.UnrepeatedDistributedSampler can be used instead "
                "of DistributedSampler."
            )

    @property
    def use_gpu(self) -> bool:
        """
        Returns True if a GPU is available, and the self.max_num_gpus flag allows it to be used. Returns False
        otherwise (i.e., if there is no GPU available, or self.max_num_gpus==0)
        """
        if self.max_num_gpus == 0:
            return False
        from health_ml.utils.common_utils import is_gpu_available

        return is_gpu_available()

    def num_gpus_per_node(self) -> int:
        """
        Computes the number of gpus to use for each node: either the number of gpus available on the device
        or restrict it to max_num_gpu, whichever is smaller. Returns 0 if running on a CPU device.
        """
        import torch

        available_gpus = torch.cuda.device_count()  # type: ignore
        num_gpus = available_gpus if self.use_gpu else 0
        message_suffix = "" if self.use_gpu else ", but not using them because use_gpu == False"
        logging.info(f"Number of available GPUs: {available_gpus}{message_suffix}")
        if 0 <= self.max_num_gpus < num_gpus:
            num_gpus = self.max_num_gpus
            logging.info(f"Restricting the number of GPUs to {num_gpus}")
        elif self.max_num_gpus > num_gpus:
            logging.warning(f"You requested max_num_gpus {self.max_num_gpus} but there are only {num_gpus} available.")
        return num_gpus



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import pickle
from io import BytesIO
from typing import Any, Optional, Callable, Dict, Union

import torch

from health_azure import RUN_CONTEXT, object_to_yaml, is_running_in_azure_ml


def _dump_to_stream(o: Any) -> BytesIO:
    """
    Writes the given object to a byte stream in pickle format, and returns the stream.

    :param o: The object to pickle.
    :return: A byte stream, with the position set to 0.
    """
    stream = BytesIO()
    pickle.dump(o, file=stream)
    stream.seek(0)
    return stream


class ModelInfo:
    """Stores a model, its example input, and metadata that describes how the model was trained."""

    MODEL = "model"
    MODEL_EXAMPLE_INPUT = "model_example_input"
    MODEL_CONFIG_YAML = "model_config_yaml"
    GIT_REPOSITORY = "git_repository"
    GIT_COMMIT_HASH = "git_commit_hash"
    DATASET_NAME = "dataset_name"
    AZURE_ML_WORKSPACE = "azure_ml_workspace"
    AZURE_ML_RUN_ID = "azure_ml_run_id"
    TEXT_TOKENIZER = "text_tokenizer"
    IMAGE_PRE_PROCESSING = "image_pre_processing"
    IMAGE_DIMENSIONS = "image_dimensions"
    OTHER_INFO = "other_info"
    OTHER_DESCRIPTION = "other_description"

    def __init__(
        self,
        model: Optional[Union[torch.nn.Module, torch.jit.ScriptModule]] = None,
        model_example_input: Optional[torch.Tensor] = None,
        model_config: Any = None,
        git_repository: str = "",
        git_commit_hash: str = "",
        dataset_name: str = "",
        azure_ml_workspace: str = "",
        azure_ml_run_id: str = "",
        text_tokenizer: Any = None,
        image_pre_processing: Optional[Callable] = None,
        image_dimensions: str = "",
        other_info: Any = None,
        other_description: str = "",
    ):
        """
        :param model: The model that should be serialized, or the deserialized model, defaults to None
        :param model_example_input: A tensor that can be input to the forward pass of the model, defaults to None
        :param model_config: The configuration object that was used to start the training run, defaults to None
        :param git_repository: The name of the git repository that contains the training codebase, defaults to ""
        :param git_commit_hash: The git commit hash that was used to run the training, defaults to ""
        :param dataset_name: The name of the dataset that was used to train the model, defaults to ""
        :param azure_ml_workspace: The name of the AzureML workspace that contains the training run, defaults to ""
        :param azure_ml_run_id: The AzureML run that did the training, defaults to ""
        :param text_tokenizer: A text tokenizer object to pre-process the model input (default: None). The object given
            here will be pickled before it is passed to ``torch.save``.
        :param image_pre_processing: An object that describes the processing for the image before it is input to the
            model, defaults to None
        :param image_dimensions: The size of the pre-processed image that is accepted by the model, defaults to ""
        :param other_info: An arbitray object that will also be written to the checkpoint. For example, this can be a
            binary stream. The object provided here will be pickled before it is passed to ``torch.save``.
        :param other_description: A human-readable description of what the ``other_info`` field contains.
        """
        self.model = model
        self.model_example_input = model_example_input
        self.model_config = model_config
        self.git_repository = git_repository
        self.git_commit_hash = git_commit_hash
        self.dataset_name = dataset_name
        self.azure_ml_workspace = azure_ml_workspace
        self.azure_ml_run_id = azure_ml_run_id
        self.text_tokenizer = text_tokenizer
        self.image_pre_processing = image_pre_processing
        self.image_dimensions = image_dimensions
        self.other_info = other_info
        self.other_description = other_description

    def get_metadata_from_azureml(self) -> None:
        """Reads information about the git repository and AzureML-related info from the AzureML run context.
        If any of those information are already stored in the object, those have higher priority.
        """
        if not is_running_in_azure_ml():
            return
        if not self.azure_ml_workspace:
            self.azure_ml_workspace = RUN_CONTEXT.experiment.workspace.name
        if not self.azure_ml_run_id:
            self.azure_ml_run_id = RUN_CONTEXT.id
        properties: Dict = RUN_CONTEXT.get_properties()
        if not self.git_repository:
            self.git_repository = properties.get("azureml.git.repository_uri", "")
        if not self.git_commit_hash:
            self.git_commit_hash = properties.get("azureml.git.commit", "")

    def state_dict(self, strict: bool = True) -> Dict[str, Any]:
        """Creates a dictionary representation of the current object.

        :param strict: The setting for the 'strict' flag in the call to torch.jit.trace.
        """

        def bytes_or_none(o: Any) -> Optional[bytes]:
            return _dump_to_stream(o).getvalue() if o is not None else None

        if self.model is None or self.model_example_input is None:
            raise ValueError("To generate a state dict, the model and model_example_input must be present.")
        try:
            traced_model = torch.jit.trace(self.model, self.model_example_input, strict=strict)
        except Exception as ex:
            raise ValueError(f"Unable to convert the model to torchscript: {ex}")
        jit_stream = BytesIO()
        torch.jit.save(traced_model, jit_stream)
        try:
            config_yaml = object_to_yaml(self.model_config) if self.model_config is not None else None
        except Exception as ex:
            raise ValueError(f"Unable to convert the model configuration to YAML: {ex}")
        return {
            ModelInfo.MODEL: jit_stream.getvalue(),
            ModelInfo.MODEL_EXAMPLE_INPUT: self.model_example_input,
            ModelInfo.MODEL_CONFIG_YAML: config_yaml,
            ModelInfo.GIT_REPOSITORY: self.git_repository,
            ModelInfo.GIT_COMMIT_HASH: self.git_commit_hash,
            ModelInfo.DATASET_NAME: self.dataset_name,
            ModelInfo.AZURE_ML_WORKSPACE: self.azure_ml_workspace,
            ModelInfo.AZURE_ML_RUN_ID: self.azure_ml_run_id,
            ModelInfo.TEXT_TOKENIZER: bytes_or_none(self.text_tokenizer),
            ModelInfo.IMAGE_PRE_PROCESSING: bytes_or_none(self.image_pre_processing),
            ModelInfo.IMAGE_DIMENSIONS: self.image_dimensions,
            ModelInfo.OTHER_INFO: bytes_or_none(self.other_info),
            ModelInfo.OTHER_DESCRIPTION: self.other_description,
        }

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """Loads a dictionary representation into the current object, overwriting all matching fields.

        :param state_dict: The dictionary to load from.
        """

        def unpickle_from_bytes(field: str) -> Any:
            if field not in state_dict:
                raise KeyError(f"State_dict does not contain a field '{field}'")
            b = state_dict[field]
            if b is None:
                return None
            stream = BytesIO(b)
            try:
                o = pickle.load(stream)
            except Exception as ex:
                raise ValueError(f"Failure when unpickling field '{field}': {ex}")
            return o

        self.model = torch.jit.load(BytesIO(state_dict[ModelInfo.MODEL]))
        self.model_example_input = state_dict[ModelInfo.MODEL_EXAMPLE_INPUT]
        self.model_config = state_dict[ModelInfo.MODEL_CONFIG_YAML]
        self.git_repository = state_dict[ModelInfo.GIT_REPOSITORY]
        self.git_commit_hash = state_dict[ModelInfo.GIT_COMMIT_HASH]
        self.dataset_name = state_dict[ModelInfo.DATASET_NAME]
        self.azure_ml_workspace = state_dict[ModelInfo.AZURE_ML_WORKSPACE]
        self.azure_ml_run_id = state_dict[ModelInfo.AZURE_ML_RUN_ID]
        self.text_tokenizer = unpickle_from_bytes(ModelInfo.TEXT_TOKENIZER)
        self.image_pre_processing = unpickle_from_bytes(ModelInfo.IMAGE_PRE_PROCESSING)
        self.image_dimensions = state_dict[ModelInfo.IMAGE_DIMENSIONS]
        self.other_info = unpickle_from_bytes(ModelInfo.OTHER_INFO)
        self.other_description = state_dict[ModelInfo.OTHER_DESCRIPTION]



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from health_ml.training_runner import TrainingRunner
from health_ml.runner import Runner


__all__ = ["TrainingRunner", "Runner"]



#  -------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  -------------------------------------------------------------------------------------------
import json
import logging
import os
import shutil
import tempfile
from math import isclose
from pathlib import Path
from typing import Any, Dict, List, Optional

from azureml.core import Run
import pandas as pd

from health_azure.utils import PARENT_RUN_CONTEXT, RUN_CONTEXT, is_running_in_azure_ml


REGRESSION_TEST_OUTPUT_FOLDER = "OUTPUT"
REGRESSION_TEST_AZUREML_FOLDER = "AZUREML_OUTPUT"
REGRESSION_TEST_AZUREML_PARENT_FOLDER = "AZUREML_PARENT_OUTPUT"
CONTENTS_MISMATCH = "Contents mismatch"
FILE_FORMAT_ERROR = "File format error"
MISSING_FILE = "Missing"
CSV_SUFFIX = ".csv"
TEXT_FILE_SUFFIXES = [".txt", ".json", ".html", ".md"]
REGRESSION_TEST_METRICS_FILENAME = "regression_metrics.json"


def compare_dictionaries(expected: Dict[str, Any], actual: Dict[str, Any], tolerance: float = 1e-5) -> None:
    """
    Function to give more clarity on the difference between two dictionaries.

    :param expected: The first dictionary to compare
    :param actual: The second dictionary to compare
    :param tolerance: The tolerance to allow when comparing numeric values, defaults to 1e-5
    """

    def _check_values_match(expected_v: Any, actual_v: Any, tolerance: float = 1e-5) -> None:
        if type(actual_v) in [float, int] and type(expected_v) in [float, int]:
            if not isclose(actual_v, expected_v, rel_tol=tolerance):
                logging.warning(f"Expected: {expected_v} does not match actual {actual_v}")
        else:
            if expected_v != actual_v:
                logging.warning(f"Expected: {expected_v} does not match actual {actual_v}")

    for expected_key, expected_val in expected.items():
        if expected_key not in actual:
            logging.warning(f"Key {expected_key} is expected but not found in actual dictionary: {actual}")
        else:
            actual_val = actual[expected_key]
            expected_type = type(expected_val)
            actual_type = type(actual_val)
            if expected_type is not actual_type:
                logging.warning(f"Actual value has type {actual_type} but we expected {expected_type}")
            if actual_type in [float, int]:
                _check_values_match(expected_val, actual_val, tolerance=tolerance)
            elif actual_type is dict:
                compare_dictionaries(expected_val, actual_val, tolerance=tolerance)
            elif actual_type in [list, set, str]:
                expected_len = len(expected_val)
                actual_len = len(actual_val)
                if expected_len != actual_len:
                    logging.warning(f"Expected value to have length {expected_len} but found {actual_len}")
                for expected_value, actual_value in zip(expected_val, actual_val):
                    _check_values_match(expected_value, actual_value, tolerance=tolerance)


def _compare_metric_values(expected: Any, actual: Any, tolerance: float = 1e-5) -> str:
    """Compares an expected and an actual value coming from a metrics dictionary. If the values are numeric
    and are close enough (within the given tolerance), an empty string is returned. If the actual
    and expected value are not matching, a string description of the discrepancy will be returned.

    :param expected: The expected value
    :param actual: The actual value
    :param tolerance: For numerical values, this is the relative tolerance allowed, defaults to 1e-5
    :return: An empty string if the expected and actual values match, otherwise a string describing the discrepancy.
    """
    if type(expected) in [float, int]:
        if type(actual) in [float, int]:
            if isclose(actual, expected, rel_tol=tolerance):
                return ""
            return f"Expected {expected} but got {actual} (allowed tolerance {tolerance})"
        return f"Expected a numeric value, but got type {type(actual).__name__}"
    return f"Don't know how to handle expected value of type {type(expected).__name__}"


def _compare_metrics_list(expected: List, actual: List, tolerance: float = 1e-5) -> List[str]:
    """Compares two lists of expected and actual values, coming from a metrics dictionary. The two lists are
    considered matching if they have the same length, and matching list elements are within the given relative
    tolerance.

    :param expected: A list of expected values.
    :param actual: A list with actual metric values.
    :param tolerance: For numerical values, this is the relative tolerance allowed, defaults to 1e-5
    :return: A list of strings describing the discrepancies between the two lists. If the lists match, the
        return value is an empty list.
    """
    messages = []
    expected_len = len(expected)
    actual_len = len(actual)
    if expected_len == actual_len:
        for index, (expected_value, actual_value) in enumerate(zip(expected, actual)):
            mismatch = _compare_metric_values(expected_value, actual_value, tolerance=tolerance)
            if mismatch:
                messages.append(f"Index {index}: {mismatch}")
    else:
        messages.append(f"Expected list of length {expected_len} but got {actual_len}")
    return messages


def compare_metrics_dictionaries(expected: Dict[str, Any], actual: Dict[str, Any], tolerance: float = 1e-5) -> str:
    """
    Function to compare two dictionaries that are expected to contain metrics (scalars or lists of scalars) or strings.
    Discrepancies are logged via logging.warning. The function returns an empty string if the dictionaries are
    identical, otherwise a string giving a short summary of the discrepancies.

    :param expected: The first dictionary to compare
    :param actual: The second dictionary to compare
    :param tolerance: The tolerance to allow when comparing numeric values, defaults to 1e-5.
    :return: An empty string if the dictionaries match, otherwise a string describing the discrepancies.
    """

    allowed_types = [float, int, str, list]
    discrepancies = 0
    for key, expected_val in expected.items():
        messages = []
        expected_type = type(expected_val)
        if expected_type not in allowed_types:
            raise ValueError(f"Expected value has type {expected_type.__name__} which is not handled.")
        if key not in actual:
            messages.append("No data found in actual metrics.")
        else:
            actual_val = actual[key]
            actual_type = type(actual_val)
            if actual_type not in allowed_types:
                messages.append(f"Actual value has type {actual_type.__name__} which is not handled.")
            elif expected_type is not actual_type:
                messages.append(
                    f"Actual value has type {actual_type.__name__} but we expected " f"{expected_type.__name__}."
                )
            elif expected_type is list:
                messages.extend(_compare_metrics_list(expected_val, actual_val, tolerance=tolerance))
            else:
                mismatch = _compare_metric_values(expected_val, actual_val, tolerance=tolerance)
                if mismatch:
                    messages.append(mismatch)
        for message in messages[:5]:
            logging.warning(f"Metric '{key}': {message}")
        if len(messages) > 0:
            discrepancies += 1
    return f"Mismatch for {discrepancies} out of {len(expected)} metrics" if discrepancies > 0 else ""


def _load_json_dict(path: Path) -> Dict[str, Any]:
    """Loads the contents of a JSON file. If the contents is a dictionary, this dictionary is returned.

    :param path: The file to load.
    :raises ValueError: If the file does not contain a JSON dictionary.
    :return: The contents of the JSON file.
    """
    metrics_json = json.loads(path.read_text())
    if not isinstance(metrics_json, Dict):
        raise ValueError(
            f"Expected metrics file {path} to contain a JSON dictionary, but got {type(metrics_json).__name__}"
        )
    return metrics_json


def _is_nested_dict(d: Dict[str, Any], message: str) -> bool:
    """Checks if a dictionary contains only dictionaries as values. Returns `True` if the dictionary contains
    only dictionaries as values. Returns `False` if none of the dictionary values are dictionaries, or if the
    dictionary is empty.
    Raises an exception if the dictionary contains a mix of dictionaries and non-dictionaries.

    :param d: The dictionary to check.
    :param message: A message to include in the exception if the dictionary contains a mix of dictionaries.
    :raises ValueError: If the dictionary contains a mix of dictionaries and non-dictionaries.
    :return: True if the dictionary contains values, and all those are dictionaries in turn. False otherwise.
    """
    has_dict_value = [isinstance(v, Dict) for v in d.values()]
    if len(d) == 0:
        return False
    if all(has_dict_value):
        return True
    if any(has_dict_value):
        raise ValueError(
            f"{message}: Metrics file has inconsistent type. Either all or none of the dictionary"
            "values should be dictionaries"
        )
    return False


def compare_metrics_files(expected: Path, actual: Path, tolerance: float = 1e-5) -> str:
    """Reads two files that contain a JSON representation of expected and actual metrics. The two files are read
    and the metrics are compared using compare_metrics_dictionaries. The function returns an empty string if the
    metrics match, otherwise a string summarizing the discrepancies. Details about the discrepancies are logged
    to `logging.warning`.

    :param expected: A file that contains the expected metrics, as a JSON dictionary.
    :param actual: A file that contains the actual metrics, as a JSON dictionary.
    :param tolerance: The maximum allowed tolerance for comparing metrics, defaults to 1e-5
    :return: An empty string if the metrics match, otherwise a string summarizing the discrepancies.
    """
    try:
        expected_metrics = _load_json_dict(expected)
        actual_metrics = _load_json_dict(actual)
        if _is_nested_dict(expected_metrics, "Expected metrics"):
            # For runs with child runs, loop over the child run dictionaries and compare them.
            if not _is_nested_dict(actual_metrics, "Actual metrics"):
                return "Expected a nested dictionary as the actual metrics, but got a flat dictionary"
            messages = []
            for key, expected_child in expected_metrics.items():
                prefix = f"Child run '{key}'"
                if key in actual_metrics:
                    actual_child = actual_metrics[key]
                    message = compare_metrics_dictionaries(expected_child, actual_child, tolerance=tolerance)
                else:
                    message = "Missing from the actual metrics"
                if message:
                    full_message = f"{prefix}: {message}"
                    logging.warning(full_message)
                    messages.append(message)
            if len(messages) > 0:
                return f"Mismatches for {len(messages)} child runs"
            return ""
        else:
            # For simple runs without child runs: Compare the dictionaries.
            if _is_nested_dict(actual_metrics, "Actual metrics"):
                return "Expected a flat dictionary as the actual metrics, but got a nested dictionary"
            return compare_metrics_dictionaries(expected_metrics, actual_metrics, tolerance=tolerance)
    except Exception as e:
        # This handles cases where the files cannot be found, are not JSON, or have inconsistent information.
        return f"Error comparing metrics files: {e}"


def compare_files(expected: Path, actual: Path, csv_relative_tolerance: float = 0.0) -> str:
    """
    Compares two individual files for regression testing. It returns an empty string if the two files appear identical.
    If the files are not identical, a brief error message is returned. Details about the mismatch are logged via
    logging.warning. This method handles CSV files (which are treated and compared as dataframes) and text files (TXT,
    JSON, HTML, MD, which are all compared while ignoring linebreaks. All other extensions are treated as binary,
    and compared on a byte-by-byte basis.

    :param expected: A file that contains the expected contents. The type of comparison (text or binary) is chosen
    based on the extension of this file.
    :param actual: A file to be checked against the expected file.
    :param csv_relative_tolerance: When comparing CSV files, maximum allowed relative discrepancy.
    If 0.0, do not allow any discrepancy.
    :return: An empty string if the files appear identical, or otherwise a brief error message. If there is a
    mismatch, details about the mismatch are printed via logging.warning.
    """

    def print_lines(prefix: str, lines: List[str]) -> None:
        num_lines = len(lines)
        count = min(5, num_lines)
        logging.warning(f"{prefix} {num_lines} lines, first {count} of those:")
        logging.warning(os.linesep.join(lines[:count]))

    def try_read_csv(prefix: str, file: Path) -> Optional[pd.DataFrame]:
        try:
            return pd.read_csv(file)
        except Exception as ex:
            logging.warning(f"{prefix} file can't be read as CSV: {str(ex)}")
            return None

    def _load_json_from_text_lines(lines: List[str]) -> Dict[str, Any]:
        return json.loads('\n'.join(lines))

    if expected.suffix == CSV_SUFFIX:
        expected_df = try_read_csv("Expected", expected)
        actual_df = try_read_csv("Actual", actual)
        if expected_df is None or actual_df is None:
            return FILE_FORMAT_ERROR
        try:
            pd.testing.assert_frame_equal(actual_df, expected_df, rtol=csv_relative_tolerance)
        except Exception as ex:
            logging.warning(str(ex))
            return CONTENTS_MISMATCH
    elif expected.suffix in TEXT_FILE_SUFFIXES:
        # Compare line-by-line to avoid issues with line separators
        expected_lines = expected.read_text().splitlines()
        actual_lines = actual.read_text().splitlines()
        if expected_lines != actual_lines:
            print_lines("Expected", expected_lines)
            print_lines("Actual", actual_lines)
            # Add additional context for json file mismatches
            if expected.suffix == ".json":
                compare_dictionaries(
                    _load_json_from_text_lines(expected_lines),
                    _load_json_from_text_lines(actual_lines),
                    tolerance=csv_relative_tolerance,
                )
            return CONTENTS_MISMATCH
    else:
        expected_binary = expected.read_bytes()
        actual_binary = actual.read_bytes()
        if expected_binary != actual_binary:
            logging.warning(f"Expected {len(expected_binary)} bytes, actual {len(actual_binary)} bytes")
            return CONTENTS_MISMATCH
    return ""


def compare_folder_contents(
    expected_folder: Path,
    actual_folder: Optional[Path] = None,
    run: Optional[Run] = None,
    csv_relative_tolerance: float = 0.0,
) -> List[str]:
    """
    Compares a set of files in a folder, against files in either the other folder or files stored in the given
    AzureML run. Each file that is present in the ``expected`` folder must be also present in the ``actual`` folder
    (or the AzureML run), with exactly the same contents, in the same folder structure.
    For example, if there is a file ``<expected>/foo/bar/contents.txt``, then there must also be a file
    ``<actual>/foo/bar/contents.txt``. If ``actual_folder`` is provided, then this is used to compare files against the
    set file files in ``expected_folder``, irrespective of the value in ``run``. If ``run`` is provided, the files
    uploaded to the AzureML run are compared against files in ``expected_folder``. If neither ``run`` nor
    ``actual_folder`` are provided, a :exc:`ValueError` is raised.

    :param expected_folder: A folder with files that are expected to be present.
    :param actual_folder: The output folder with the actually produced files.
    :param run: An AzureML run
    :param csv_relative_tolerance: When comparing CSV files, use this as the maximum allowed relative discrepancy.
    If 0.0, do not allow any discrepancy.
    :return: A list of human readable error messages, with message and file path. If no errors are found, the list is
    empty.
    """
    messages = []
    if run and not is_running_in_azure_ml(run):
        logging.warning("Skipping file comparison because the given run context is an AzureML offline run")
        return []
    files_in_run: List[str] = run.get_file_names() if run else []
    temp_folder = Path(tempfile.mkdtemp()) if run else None
    for file in expected_folder.rglob("*"):
        # rglob also returns folders, skip those
        if file.is_dir():
            continue
        # All files stored in AzureML runs use Linux-style path
        file_relative = file.relative_to(expected_folder).as_posix()
        if actual_folder:
            actual_file = actual_folder / file_relative
        elif temp_folder is not None and run is not None:
            actual_file = temp_folder / file_relative
            if file_relative in files_in_run:
                run.download_file(name=str(file_relative), output_file_path=str(actual_file))
        else:
            raise ValueError("Either of the two arguments 'run' or 'actual_folder' must be provided")
        if file_relative == REGRESSION_TEST_METRICS_FILENAME:
            message = compare_metrics_files(expected=file, actual=actual_file, tolerance=csv_relative_tolerance)
        elif actual_file.exists():
            message = compare_files(expected=file, actual=actual_file, csv_relative_tolerance=csv_relative_tolerance)
        else:
            message = MISSING_FILE
        if message:
            messages.append(f"{message}: {file_relative}")
            logging.warning(f"File {file_relative}: {message}")
        else:
            logging.info(f"File {file_relative}: OK")
    if temp_folder:
        shutil.rmtree(temp_folder)
    return messages


def compare_folders_and_run_outputs(expected: Path, actual: Path, csv_relative_tolerance: float = 0.0) -> None:
    """
    Compares the actual set of run outputs in the ``actual`` folder against an expected set of files in the ``expected``
    folder. The ``expected`` folder can have two special subfolders AZUREML_OUTPUT and AZUREML_PARENT_OUTPUT, that
    contain files that are expected to be present in the AzureML run context of the present run (AZUREML_OUTPUT)
    or the run context of the parent run (AZUREML_PARENT_OUTPUT).
    If a file is missing, or does not have the expected contents, an exception is raised.

    :param expected: A folder with files that are expected to be present.
    :param actual: The output folder with the actually produced files.
    :param csv_relative_tolerance: When comparing CSV files, use this as the maximum allowed relative discrepancy.
    If 0.0, do not allow any discrepancy.
    """
    if not expected.is_dir():
        raise ValueError(f"Folder with expected files does not exist: {expected}")
    logging.debug(f"Current working directory: {Path.cwd()}")
    messages = []
    folders_to_check = [
        (REGRESSION_TEST_OUTPUT_FOLDER, "run output files", actual, None),
        (REGRESSION_TEST_AZUREML_FOLDER, "AzureML outputs in present run", None, RUN_CONTEXT),
        (REGRESSION_TEST_AZUREML_PARENT_FOLDER, "AzureML outputs in parent run", None, PARENT_RUN_CONTEXT),
    ]
    for subfolder, message_prefix, actual_folder, run_to_compare in folders_to_check:
        folder = expected / subfolder
        if folder.is_dir():
            logging.info(f"Comparing results in {folder} against {message_prefix}:")
            if actual_folder is None and run_to_compare is None:
                logging.info("No AzureML run to compare against. Skipping")
                continue
            new_messages = compare_folder_contents(
                folder, actual_folder=actual_folder, run=run_to_compare, csv_relative_tolerance=csv_relative_tolerance
            )
            if new_messages:
                messages.append(f"Issues in {message_prefix}:")
                messages.extend(new_messages)
        else:
            logging.info(f"Folder {subfolder} not found, skipping comparison against {message_prefix}")
    if messages:
        raise ValueError(
            f"Some expected files were missing or did not have the expected contents:{os.linesep}"
            f"{os.linesep.join(messages)}"
        )



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

import os
from contextlib import contextmanager
from datetime import datetime
from enum import Enum, unique
from pathlib import Path
from typing import Any, Generator, List, Optional

import torch
from torch.nn import Module
import pandas as pd
from health_azure import paths

from health_azure.utils import PathOrString, is_conda_file_with_pip_include, find_file_in_parent_folders

MAX_PATH_LENGTH = 260

# convert string to None if an empty string or whitespace is provided


def empty_string_to_none(x: Optional[str]) -> Optional[str]:
    return None if (x is None or len(x.strip()) == 0) else x


def string_to_path(x: Optional[str]) -> Optional[Path]:
    return None if (x is None or len(x.strip()) == 0) else Path(x)


# file and directory names
CHECKPOINT_SUFFIX = ".ckpt"
AUTOSAVE_CHECKPOINT_FILE_NAME = "autosave"
AUTOSAVE_CHECKPOINT_CANDIDATES = [
    AUTOSAVE_CHECKPOINT_FILE_NAME + CHECKPOINT_SUFFIX,
    AUTOSAVE_CHECKPOINT_FILE_NAME + "-v1" + CHECKPOINT_SUFFIX,
]
CHECKPOINT_FOLDER = "checkpoints"
DEFAULT_AML_UPLOAD_DIR = "outputs"
DEFAULT_LOGS_DIR_NAME = "logs"
EXPERIMENT_SUMMARY_FILE = "experiment_summary.txt"
DEFAULT_AML_CHECKPOINT_DIR = f"{DEFAULT_AML_UPLOAD_DIR}/{CHECKPOINT_FOLDER}"

# run recovery
RUN_RECOVERY_ID_KEY = "run_recovery_id"
RUN_RECOVERY_FROM_ID_KEY_NAME = "recovered_from"

# other
EFFECTIVE_RANDOM_SEED_KEY_NAME = "effective_random_seed"


@unique
class ModelExecutionMode(Enum):
    """
    Model execution mode
    """

    TRAIN = "Train"
    TEST = "Test"
    VAL = "Val"


def is_windows() -> bool:
    """
    Returns True if the host operating system is Windows.
    """
    return os.name == "nt"


def is_linux() -> bool:
    """
    Returns True if the host operating system is a flavour of Linux.
    """
    return os.name == "posix"


def check_properties_are_not_none(obj: Any, ignore: Optional[List[str]] = None) -> None:
    """
    Checks to make sure the provided object has no properties that have a None value assigned.
    """
    if ignore is not None:
        none_props = [k for k, v in vars(obj).items() if v is None and k not in ignore]
        if len(none_props) > 0:
            raise ValueError("Properties had None value: {}".format(none_props))


@contextmanager
def change_working_directory(path_or_str: PathOrString) -> Generator:
    """
    Context manager for changing the current working directory to the value provided. Outside the context
    manager, the original working directory will be restored.

    :param path_or_str: The new directory to change to
    :yield: a _GeneratorContextManager object (this object itself is of no use, rather we are interested in
        the side effect of the working directory temporarily changing
    """
    new_path = Path(path_or_str).expanduser()
    if not new_path.is_dir():
        raise FileNotFoundError(f"Path {new_path} is not a directory")
    old_path = Path.cwd()
    os.chdir(new_path)
    yield
    os.chdir(str(old_path))


def _create_generator(seed: Optional[int] = None) -> torch.Generator:
    """
    Create Torch generator and sets seed with value if provided, or else with a random seed.

    :param seed: Optional seed to set the Generator object with
    :return: Torch Generator object
    """
    generator = torch.Generator()
    if seed is None:
        seed = int(torch.empty((), dtype=torch.int64).random_().item())
    generator.manual_seed(seed)
    return generator


def choose_conda_env_file(env_file: Optional[Path] = None) -> Path:
    """
    Chooses the Conda environment file that should be used when submitting the present run to AzureML. If a Conda
    file is given explicitly on the commandline, return that. Otherwise, look in the current folder and its parents for
    a file called `environment.yml`.

    :param env_file: The Conda environment file that was specified on the commandline when starting the run.
    :return: The Conda environment files to use.
    :raises FileNotFoundError: If the specified Conda file does not exist, or none could be found at all.
    """
    if env_file is not None:
        if env_file.is_file():
            return env_file
        raise FileNotFoundError(f"The Conda file specified on the commandline could not be found: {env_file}")
    # When running from the Git repo, then stop search for environment file at repository root. Otherwise,
    # search from current folder all the way up
    stop_at = [paths.git_repo_root_folder()] if paths.is_himl_used_from_git_repo() else []
    current_folder = Path.cwd()
    env_file = find_file_in_parent_folders(
        paths.ENVIRONMENT_YAML_FILE_NAME, start_at_path=current_folder, stop_at_path=stop_at
    )
    if env_file is None:
        raise FileNotFoundError(
            f"No Conda environment file '{paths.ENVIRONMENT_YAML_FILE_NAME}' was found in the "
            f"current folder {current_folder} or its parent folders"
        )
    return env_file


def check_conda_environment(env_file: Path) -> None:
    """Tests if the given conda environment files is valid. In particular, it must not contain "include" statements
    in the pip section.

    :param env_file: The Conda environment YAML file to check.
    """
    has_pip_include, _ = is_conda_file_with_pip_include(env_file)
    if has_pip_include:
        raise ValueError(
            f"The Conda environment definition in {env_file} uses '-r' to reference pip requirements "
            "files. This does not work in AzureML. Please add the pip dependencies directly."
        )


def create_unique_timestamp_id() -> str:
    """
    Creates a unique string using the current time in UTC, up to seconds precision, with characters that
    are suitable for use in filenames. For example, on 31 Dec 2019 at 11:59:59pm UTC, the result would be
    2019-12-31T235959Z.
    """
    unique_id = datetime.utcnow().strftime("%Y-%m-%dT%H%M%SZ")
    return unique_id


def is_gpu_available() -> bool:
    """

    :return: True if a GPU with at least 1 device is available.
    """
    return torch.cuda.is_available() and torch.cuda.device_count() > 0  # type: ignore


@contextmanager
def set_model_to_eval_mode(model: Module) -> Generator:
    """
    Puts the given torch model into eval mode. At the end of the context, resets the state of the training flag to
    what is was before the call.

    :param model: The model to modify.
    """
    old_mode = model.training
    model.eval()
    yield
    model.train(old_mode)


def is_long_path(path: PathOrString) -> bool:
    """
    A long path is a path that has more than MAX_PATH_LENGTH characters

    :param path: The path to check the length of
    :return: True if the length of the path is greater than MAX_PATH_LENGTH, else False
    """
    return len(str(path)) > MAX_PATH_LENGTH


def df_to_json(df: pd.DataFrame, json_path: Path, add_newline: bool = True) -> None:
    """Save a data frame to a JSON file.

    :param df: Input data frame.
    :param json_path: Path to output JSON file.
    :param add_newline: If ``True``, add newline at the end of the JSON file for POSIX compliance.
    """
    text = df.to_json()
    if add_newline:
        text += '\n'
    json_path.write_text(text)


def seed_monai_if_available(seed: int) -> None:
    """If the MONAI package is available, set its shared seed to make all MONAI operations deterministic.
    If MONAI is not available, nothing will happen.

    :param seed: The random seed to use for MONAI."""
    try:
        # MONAI is not part of the core hi-ml requirements, this import can fail.
        from monai.utils import set_determinism  # type: ignore

        set_determinism(seed=seed)
    except ImportError:
        pass



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from pathlib import Path
from typing import Dict, List, Tuple, TypeVar, Union

T = TypeVar('T')
IntOrFloat = Union[int, float]
PathOrString = Union[Path, str]
TupleFloat2 = Tuple[float, float]
TupleInt3 = Tuple[int, int, int]
DictStrFloat = Dict[str, float]
DictStrFloatOrFloatList = Dict[str, Union[float, List[float]]]



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import cv2
import numpy as np
import torch
import torchvision.transforms.functional as TF


class HEDJitter(object):
    """
    Randomly perturbe the HED color space value an RGB image.

    First, it disentangled the hematoxylin and eosin color channels by color deconvolution method using a fixed matrix.
    Second, it perturbed the hematoxylin, eosin stains independently.
    Third, it transformed the resulting stains into regular RGB color space.
    PyTorch version of: https://github.com/gatsby2016/Augmentation-PyTorch-Transforms/blob/master/myTransforms.py

    Usage example:
        >>> transform = HEDJitter(0.05)
        >>> img = transform(img)
    """

    def __init__(self, theta: float = 0.0) -> None:
        """
        :param theta: How much to jitter HED color space.
            HED_light: theta=0.05; HED_strong: theta=0.2.
            alpha is chosen from a uniform distribution [1-theta, 1+theta].
            beta is chosen from a uniform distribution [-theta, theta].
            The jitter formula is :math:`s' = \alpha * s + \beta`.
        """
        self.theta = theta
        self.rgb_from_hed = torch.tensor([[0.65, 0.70, 0.29], [0.07, 0.99, 0.11], [0.27, 0.57, 0.78]])
        self.hed_from_rgb = torch.tensor(
            [
                [1.87798274, -1.00767869, -0.55611582],
                [-0.06590806, 1.13473037, -0.1355218],
                [-0.60190736, -0.48041419, 1.57358807],
            ]
        )
        self.log_adjust = torch.log(torch.tensor(1e-6))

    def adjust_hed(self, img: torch.Tensor) -> torch.Tensor:
        """
        Applies HED jitter to image.

        :param img: Input image.
        """

        alpha = torch.FloatTensor(img.shape[0], 1, 1, 3).uniform_(1 - self.theta, 1 + self.theta)
        beta = torch.FloatTensor(img.shape[0], 1, 1, 3).uniform_(-self.theta, self.theta)

        # Separate stains
        img = img.permute([0, 2, 3, 1])
        img = torch.maximum(img, 1e-6 * torch.ones(img.shape))
        stains = (torch.log(img) / self.log_adjust) @ self.hed_from_rgb
        stains = torch.maximum(stains, torch.zeros(stains.shape))

        # perturbations in HED color space
        stains = alpha * stains + beta

        # Combine stains
        img = -(stains * (-self.log_adjust)) @ self.rgb_from_hed
        img = torch.exp(img)
        img = torch.clip(img, 0, 1)
        img = img.permute(0, 3, 1, 2)

        # Normalize
        imin = torch.amin(img, dim=[1, 2, 3], keepdim=True)
        imax = torch.amax(img, dim=[1, 2, 3], keepdim=True)
        img = (img - imin) / (imax - imin)

        return img

    def __call__(self, img: torch.Tensor) -> torch.Tensor:
        original_shape = img.shape
        if len(original_shape) == 3:
            img = img.unsqueeze(0)  # add batch dimension if missing
        # if the input is a bag of images, hed jitter needs to run on each image separately
        if img.shape[0] > 1:
            for i in range(img.shape[0]):
                img_tile = img[i]
                img[i] = self.adjust_hed(img_tile.unsqueeze(0))
            return img
        else:
            img = self.adjust_hed(img)
            if len(original_shape) == 3:
                return img.squeeze(0)
            return img


class StainNormalization(object):
    """Normalize the stain of an image given a reference image.

    Following Erik Reinhard, Bruce Gooch (2001): “Color Transfer between Images.”
    First, mask all white pixels.
    Second, convert remaining pixels to lab space and normalize each channel.
    Third, add mean and std of reference image.
    Fourth, convert back to rgb and add white pixels back.

    Usage example:
        >>> transform = StainNormalization()
        >>> img = transform(img)
    """

    def __init__(self) -> None:
        # mean and std per channel of a reference image
        self.reference_mean = np.array([148.60, 169.30, 105.97])
        self.reference_std = np.array([41.56, 9.01, 6.67])

    @staticmethod
    def stain_normalize(img: torch.Tensor, reference_mean: np.ndarray, reference_std: np.ndarray) -> torch.Tensor:
        """
        Applies stain normalization to image.

        :param img: Input image.
        :param reference_mean: Mean per channel of a reference image.
        :param reference_std: STD per channel of a reference image.
        """
        img = img.permute([0, 2, 3, 1]).squeeze().numpy() * 255  # only 3 channels, color channel last, range 0 - 255
        img = img.astype(np.uint8)  # type: ignore

        whitemask = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
        whitemask = whitemask > 215
        whitemask = np.repeat(whitemask[:, :, np.newaxis], 3, axis=2)

        imagelab = cv2.cvtColor(img, cv2.COLOR_RGB2LAB)
        imagelab_masked = np.ma.MaskedArray(imagelab, whitemask)  # type: np.ma.MaskedArray

        epsilon = 1e-11  # Sometimes STD is near 0, add epsilon to avoid div by 0
        imagelab_masked_mean = imagelab_masked.mean(axis=(0, 1))
        imagelab_masked_std = imagelab_masked.std(axis=(0, 1)) + epsilon

        # Normalize and apply reference img statistics
        imagelab = (imagelab - imagelab_masked_mean) / imagelab_masked_std * reference_std + reference_mean
        imagelab = np.clip(imagelab, 0, 255)
        imagelab = imagelab.astype(np.uint8)
        nimg = cv2.cvtColor(imagelab, cv2.COLOR_LAB2RGB)
        nimg[whitemask] = img[whitemask]  # add back white pixels
        nimg = torch.Tensor(nimg).unsqueeze(0).permute(0, 3, 1, 2) / 255.0  # back to pytorch format

        return nimg

    def __call__(self, img: torch.Tensor) -> torch.Tensor:
        original_shape = img.shape
        if len(original_shape) == 3:
            img = img.unsqueeze(0)  # add batch dimension if missing
        # if the input is a bag of images, stain normalization needs to run on each image separately
        if img.shape[0] > 1:
            for i in range(img.shape[0]):
                img_tile = img[i]
                img[i] = self.stain_normalize(img_tile.unsqueeze(0), self.reference_mean, self.reference_std)
            return img
        else:
            img = self.stain_normalize(img, self.reference_mean, self.reference_std)
            if len(original_shape) == 3:
                return img.squeeze(0)
            return img


class GaussianBlur(object):
    """
    Implements Gaussian blur as described in the SimCLR paper (https://arxiv.org/abs/2002.05709).

    Blur image using a Gaussian kernel with a randomly sampled STD.
    Slight modification of the code in pl_bolts to make it work with our transform pipeline.

    Usage example:
            >>> transform = GaussianBlur(kernel_size=int(224 * 0.1) + 1)
            >>> img = transform(img)
    """

    def __init__(self, kernel_size: int, p: float = 0.5, min: float = 0.1, max: float = 2.0) -> None:
        """
        :param kernel_size: Size of the Gaussian kernel, e.g., about 10% of the image size.
        :param p: Probability of applying blur.
        :param min: lower bound of the interval from which we sample the STD
        :param max: upper bound of the interval from which we sample the STD
        """
        self.min = min
        self.max = max
        self.kernel_size = kernel_size
        self.p = p

    @staticmethod
    def apply_gaussian_blur(sample: torch.Tensor, kernel_size: int, p: float, min: float, max: float) -> torch.Tensor:
        """
        Applies Gaussian blur to image.

        :param img: Input image.
        :param kernel_size: Size of the Gaussian kernel, e.g., about 10% of the image size.
        :param p: Probability of applying blur.
        :param min: lower bound of the interval from which we sample the STD
        :param max: upper bound of the interval from which we sample the STD
        """
        if np.random.binomial(n=1, p=p):
            sigma = np.random.uniform(low=min, high=max)  # (max - min) * np.random.random_sample() + min
            sample = sample.permute([0, 2, 3, 1]).squeeze().numpy()  # only 3 channels, color channel last
            sample = cv2.GaussianBlur(sample, (kernel_size, kernel_size), sigma)
            sample = torch.Tensor(sample).unsqueeze(0).permute(0, 3, 1, 2)
        return sample

    def __call__(self, img: torch.Tensor) -> torch.Tensor:
        original_shape = img.shape
        if len(original_shape) == 3:
            img = img.unsqueeze(0)  # add batch dimension if missing
        # if the input is a bag of images, gaussian blur needs to run on each image separately
        if img.shape[0] > 1:
            for i in range(img.shape[0]):
                img_tile = img[i]
                img[i] = self.apply_gaussian_blur(
                    sample=img_tile.unsqueeze(0), kernel_size=self.kernel_size, p=self.p, min=self.min, max=self.max
                )
            return img
        else:
            img = self.apply_gaussian_blur(
                sample=img, kernel_size=self.kernel_size, p=self.p, min=self.min, max=self.max
            )
            if len(original_shape) == 3:
                return img.squeeze(0)
            return img


class RandomRotationByMultiplesOf90(object):
    """
    Rotation of input image by 0, 90, 180 or 270 degrees.

    Usage example:
            >>> transform = RandomRotationByMultiplesOf90()
            >>> img = transform(img)
    """

    def __init__(self) -> None:
        super().__init__()

    def __call__(self, sample: torch.Tensor) -> torch.Tensor:
        angle = np.random.choice([0.0, 90.0, 180.0, 270.0])

        if angle != 0.0:
            sample = TF.rotate(sample, angle)

        return sample



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import logging
from argparse import Namespace
from typing import Any, Dict, Iterable, List, Optional, Union

from pytorch_lightning import Trainer
from pytorch_lightning.loggers import LightningLoggerBase, MLFlowLogger
from pytorch_lightning.utilities.logger import _convert_params, _flatten_dict
from pytorch_lightning.utilities.rank_zero import rank_zero_only, rank_zero_warn
from health_ml.utils.type_annotations import DictStrFloat, DictStrFloatOrFloatList


class StoringLogger(LightningLoggerBase):
    """
    A Pytorch Lightning logger that simply stores the metrics that are written to it.
    Used for diagnostic purposes in unit tests.
    """

    def __init__(self) -> None:
        super().__init__()
        self.results_per_epoch: Dict[int, DictStrFloatOrFloatList] = {}
        self.hyperparams: Any = None
        # Fields to store diagnostics for unit testing
        self.train_diagnostics: List[Any] = []
        self.val_diagnostics: List[Any] = []
        self.results_without_epoch: List[DictStrFloat] = []

    @rank_zero_only
    def log_metrics(self, metrics: DictStrFloat, step: Optional[int] = None) -> None:
        logging.debug(f"StoringLogger step={step}: {metrics}")
        epoch_name = "epoch"
        if epoch_name not in metrics:
            # Metrics without an "epoch" key are logged during testing, for example
            self.results_without_epoch.append(metrics)
            return
        epoch = int(metrics[epoch_name])
        del metrics[epoch_name]
        for key, value in metrics.items():
            if isinstance(value, int):
                metrics[key] = float(value)
        if epoch in self.results_per_epoch:
            current_results = self.results_per_epoch[epoch]
            for key, value in metrics.items():
                if key in current_results:
                    logging.debug(f"StoringLogger: appending results for metric {key}")
                    current_metrics = current_results[key]
                    if isinstance(current_metrics, list):
                        current_metrics.append(value)
                    else:
                        current_results[key] = [current_metrics, value]
                else:
                    current_results[key] = value
        else:
            self.results_per_epoch[epoch] = metrics  # type: ignore

    @rank_zero_only
    def log_hyperparams(self, params: Any) -> None:
        self.hyperparams = params

    def experiment(self) -> Any:
        return None

    def name(self) -> Any:
        return ""

    def version(self) -> int:
        return 0

    @property
    def epochs(self) -> Iterable[int]:
        """
        Gets the epochs for which the present object holds any results.
        """
        return self.results_per_epoch.keys()

    def extract_by_prefix(self, epoch: int, prefix_filter: str = "") -> DictStrFloat:
        """
        Reads the set of metrics for a given epoch, filters them to retain only those that have the given prefix,
        and returns the filtered ones. This is used to break a set
        of results down into those for training data (prefix "Train/") or validation data (prefix "Val/").

        :param epoch: The epoch for which results should be read.
        :param prefix_filter: If empty string, return all metrics. If not empty, return only those metrics that
        have a name starting with `prefix`, and strip off the prefix.
        :return: A metrics dictionary.
        """
        epoch_results = self.results_per_epoch.get(epoch, None)
        if epoch_results is None:
            raise KeyError(f"No results are stored for epoch {epoch}")
        filtered = {}
        for key, value in epoch_results.items():
            assert isinstance(key, str), f"All dictionary keys should be strings, but got: {type(key)}"
            # Add the metric if either there is no prefix filter (prefix does not matter), or if the prefix
            # filter is supplied and really matches the metric name
            if (not prefix_filter) or key.startswith(prefix_filter):
                stripped_key = key[len(prefix_filter) :]
                filtered[stripped_key] = value  # type: ignore
        return filtered  # type: ignore

    def to_metrics_dicts(self, prefix_filter: str = "") -> Dict[int, DictStrFloat]:
        """
        Converts the results stored in the present object into a two-level dictionary, mapping from epoch number to
        metric name to metric value. Only metrics where the name starts with the given prefix are retained, and the
        prefix is stripped off in the result.

        :param prefix_filter: If empty string, return all metrics. If not empty, return only those metrics that
        have a name starting with `prefix`, and strip off the prefix.
        :return: A dictionary mapping from epoch number to metric name to metric value.
        """
        return {epoch: self.extract_by_prefix(epoch, prefix_filter) for epoch in self.epochs}


class HimlMLFlowLogger(MLFlowLogger):
    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    @rank_zero_only
    def log_hyperparams(self, params: Union[Dict[str, Any], Namespace]) -> None:
        """
        Override underlying log_hyperparams message to avoid trying to log hyperparameters that have already
        been logged, thus causing MLFlow to raise an Exception.

        :param params: The original hyperparameters to be logged.
        """
        run = self._mlflow_client.get_run(self.run_id)
        existing_hyperparams = run.data.params

        params = _convert_params(params)
        params = _flatten_dict(params)
        for k, v in params.items():
            if len(str(v)) > 250:
                rank_zero_warn(
                    f"Mlflow only allows parameters with up to 250 characters. Discard {k}={v}", category=RuntimeWarning
                )
                continue
            if k in existing_hyperparams:
                continue

            self.experiment.log_param(self.run_id, k, v)


def get_mlflow_run_id_from_trainer(trainer: Optional[Trainer]) -> Optional[str]:
    """
    If trainer has already been intialised with loggers, attempt to retrieve one of the type HimlMLFlowLogger,
    and return its run_id property in order to log to the same run. Otherwise, return None.

    :return: The mlflow run id from an existing HimlMLFlowLogger if available, else None.
    """
    if trainer is None:
        return None
    try:
        mlflow_logger = [logger for logger in trainer.loggers if isinstance(logger, HimlMLFlowLogger)][0]
        return mlflow_logger.run_id
    except IndexError:
        return None



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import torch
from pytorch_lightning import LightningDataModule, LightningModule
from pytorch_lightning.callbacks import Callback
from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint
from torchmetrics import MeanAbsoluteError
from torch.optim import Adam, Optimizer
from torch.optim.lr_scheduler import StepLR, _LRScheduler
from torch.utils.data import DataLoader, Dataset

from health_ml.lightning_container import LightningContainer

TEST_MSE_FILE = "test_mse.txt"
TEST_MAE_FILE = "test_mae.txt"


def _create_1d_regression_dataset(n: int = 100, seed: int = 0) -> torch.Tensor:
    """Creates a simple 1-D dataset of a noisy linear function.

    :param n: The number of datapoints to generate, defaults to 100
    :param seed: Random number generator seed, defaults to 0
    :return: A tensor that contains X values in [:, 0] and Y values in [:, 1]
    """
    torch.manual_seed(seed)
    x = torch.rand((n, 1)) * 10
    y = 0.2 * x + 0.1 * torch.randn(x.size())
    xy = torch.cat((x, y), dim=1)
    return xy


def _split_crossval(xy: torch.Tensor, crossval_count: int, crossval_index: int) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Generates a split of the given dataset along the first dimension for cross-validation.

    :param xy: The data that should be split. The split will be generated acros dimension 0.
    :param crossval_count: The number of splits in total
    :param crossval_index: The index of the split that should be generated (0 <= crossval_index < crossval_count)
    :return: A tuple of (training data, validation data)
    """
    n = xy.shape[0]
    split_size = n // crossval_count
    val_start = crossval_index * split_size
    val_end = (crossval_index + 1) * split_size
    train1_start = 0 if crossval_index == 0 else (crossval_index - 1) * split_size
    train1_end = 0 if crossval_index == 0 else val_start
    train2_start = val_end if crossval_index < (crossval_count - 1) else 0
    train2_end = n if crossval_index < (crossval_count - 1) else 0
    val = xy[val_start:val_end]
    train = torch.concat([xy[train1_start:train1_end], xy[train2_start:train2_end]])
    return (train, val)


class HelloWorldDataset(Dataset):
    """
    A simple 1dim regression task
    """

    def __init__(self, xy: torch.Tensor) -> None:
        """
        Creates the 1-dim regression dataset.

        :param xy: The raw data, x in the first column, y in the second column
        """
        super().__init__()  # type: ignore
        self.xy = xy

    def __len__(self) -> int:
        return self.xy.shape[0]

    def __getitem__(self, item: int) -> Dict[str, torch.Tensor]:
        return {"x": self.xy[item][0:1], "y": self.xy[item][1:2]}


class HelloWorldDataModule(LightningDataModule):
    """
    A data module that gives the training, validation and test data for a simple 1-dim regression task.
    """

    def __init__(self, crossval_count: int, crossval_index: Optional[int] = None, seed: int = 0) -> None:
        super().__init__()
        n_total = 200
        xy = _create_1d_regression_dataset(n=n_total, seed=seed)
        n_test = 40
        n_val = 50
        self.test = HelloWorldDataset(xy=xy[:n_test])
        if crossval_count <= 1:
            self.val = HelloWorldDataset(xy=xy[n_test : (n_test + n_val)])
            self.train = HelloWorldDataset(xy=xy[(n_test + n_val) :])
        else:
            # This could be done via a library function like sklearn's KFold function, but we don't want to add
            # scikit-learn as a dependency just for this example.
            assert crossval_index is not None, "crossval_index must be set if crossval_count > 1"
            train, val = _split_crossval(xy[n_test:], crossval_count=crossval_count, crossval_index=crossval_index)
            self.val = HelloWorldDataset(xy=val)
            self.train = HelloWorldDataset(xy=train)

    def prepare_data(self, *args: Any, **kwargs: Any) -> None:
        pass

    def setup(self, stage: Optional[str] = None) -> None:
        pass

    def train_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        return DataLoader(self.train, batch_size=5)

    def val_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        return DataLoader(self.val, batch_size=5)

    def test_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        return DataLoader(self.test, batch_size=5)


class HelloRegression(LightningModule):
    """
    A simple 1-dim regression model.
    """

    def __init__(self) -> None:
        super().__init__()
        self.model = torch.nn.Linear(in_features=1, out_features=1, bias=True)  # type: ignore
        self.test_mse: List[torch.Tensor] = []
        self.test_mae = MeanAbsoluteError()
        self._on_extra_val_epoch = False

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore
        """
        This method is part of the standard PyTorch Lightning interface. For an introduction, please see
        https://pytorch-lightning.readthedocs.io/en/stable/starter/converting.html
        It runs a forward pass of a tensor through the model.

        :param x: The input tensor(s)
        :return: The model output.
        """
        return self.model(x)

    def training_step(self, batch: Dict[str, torch.Tensor], *args: Any, **kwargs: Any) -> torch.Tensor:  # type: ignore
        """
        This method is part of the standard PyTorch Lightning interface. For an introduction, please see
        https://pytorch-lightning.readthedocs.io/en/stable/starter/converting.html
        It consumes a minibatch of training data (coming out of the data loader), does forward propagation, and
        computes the loss.

        :param batch: The batch of training data
        :return: The loss value with a computation graph attached.
        """
        loss = self.shared_step(batch)
        self.log("loss", loss, on_epoch=True, on_step=False)
        return loss

    def validation_step(  # type: ignore
        self, batch: Dict[str, torch.Tensor], *args: Any, **kwargs: Any
    ) -> torch.Tensor:
        """
        This method is part of the standard PyTorch Lightning interface. For an introduction, please see
        https://pytorch-lightning.readthedocs.io/en/stable/starter/converting.html
        It consumes a minibatch of validation data (coming out of the data loader), does forward propagation, and
        computes the loss.

        :param batch: The batch of validation data
        :return: The loss value on the validation data.
        """
        loss = self.shared_step(batch)
        self.log("val_loss", loss, on_epoch=True, on_step=False)
        return loss

    def shared_step(self, batch: Dict[str, torch.Tensor]) -> torch.Tensor:
        """
        This is a convenience method to reduce code duplication, because training, validation, and test step share
        large amounts of code.

        :param batch: The batch of data to process, with input data and targets.
        :return: The MSE loss that the model achieved on this batch.
        """
        input = batch["x"]
        target = batch["y"]
        prediction = self.forward(input)
        return torch.nn.functional.mse_loss(prediction, target)  # type: ignore

    def configure_optimizers(self) -> Tuple[List[Optimizer], List[_LRScheduler]]:
        """
        This method is part of the standard PyTorch Lightning interface. For an introduction, please see
        https://pytorch-lightning.readthedocs.io/en/stable/starter/converting.html
        It returns the PyTorch optimizer(s) and learning rate scheduler(s) that should be used for training.
        """
        optimizer = Adam(self.parameters(), lr=1e-1)
        scheduler = StepLR(optimizer, step_size=20, gamma=0.5)
        return [optimizer], [scheduler]

    def on_test_epoch_start(self) -> None:
        """
        This method is part of the standard PyTorch Lightning interface. For an introduction, please see
        https://pytorch-lightning.readthedocs.io/en/stable/starter/converting.html
        In this method, you can prepare data structures that need to be in place before evaluating the model on the
        test set (that is done in the test_step).
        """
        self.test_mse = []
        self.test_mae.reset()

    def test_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> torch.Tensor:  # type: ignore
        """
        This method is part of the standard PyTorch Lightning interface. For an introduction, please see
        https://pytorch-lightning.readthedocs.io/en/stable/starter/converting.html
        It evaluates the model in "inference mode" on data coming from the test set. It could, for example,
        also write each model prediction to disk.

        :param batch: The batch of test data.
        :param batch_idx: The index (0, 1, ...) of the batch when the data loader is enumerated.
        :return: The loss on the test data.
        """
        input = batch["x"]
        target = batch["y"]
        prediction = self.forward(input)
        # This illustrates two ways of computing metrics: Using standard torch
        loss = torch.nn.functional.mse_loss(prediction, target)  # type: ignore
        self.test_mse.append(loss)
        # Metrics computed using PyTorch Lightning objects. Note that these will, by default, attempt
        # to synchronize across GPUs.
        self.test_mae.update(preds=prediction, target=target)
        return loss

    def on_test_epoch_end(self) -> None:
        """
        This method is part of the standard PyTorch Lightning interface. For an introduction, please see
        https://pytorch-lightning.readthedocs.io/en/stable/starter/converting.html
        In this method, you can finish off anything to do with evaluating the model on the test set,
        for example writing aggregate metrics to disk.
        """
        average_mse = torch.mean(torch.stack(self.test_mse))
        Path(TEST_MSE_FILE).write_text(str(average_mse.item()))
        Path(TEST_MAE_FILE).write_text(str(self.test_mae.compute().item()))
        self.log("test_mse", average_mse, on_epoch=True, on_step=False)

    def on_run_extra_validation_epoch(self) -> None:
        self._on_extra_val_epoch = True


class HelloWorld(LightningContainer):
    """
    An example container for using the hi-ml runner. This container has methods
    to generate the actual Lightning model, and read out the datamodule that will be used for training.
    The number of training epochs is controlled at container level.
    You can train this model by running `python health_ml/runner.py --model=HelloWorld` on the local box,
    or via `python health_ml/runner.py --model=HelloWorld --cluster=<cluster name>` in AzureML
    """

    def __init__(self) -> None:
        super().__init__()
        self.local_dataset_dir = Path(__file__).resolve().parent
        self.max_epochs = 20
        self.save_checkpoint = False

    # This method must be overridden by any subclass of LightningContainer. It returns the model that you wish to
    # train, as a LightningModule
    def create_model(self) -> LightningModule:
        return HelloRegression()

    # This method must be overridden by any subclass of LightningContainer. It returns a data module, which
    # in turn contains 3 data loaders for training, validation, and test set.
    def get_data_module(self) -> LightningDataModule:
        assert self.local_dataset_dir is not None
        # If you would like to use the built-in cross validation functionality that runs training in parallel,
        # you need to provide the crossvalidation parameters in the LightningContainer to the datamodule. The
        # datamodule must carry out appropriate splitting of the data.
        return HelloWorldDataModule(crossval_count=self.crossval_count, crossval_index=self.crossval_index)

    # This method is optional. Override it to supply a data module to use for evaluating the model on a new dataset.
    # Only the test data loader from the data module will be used.
    def get_eval_data_module(self) -> LightningDataModule:
        return HelloWorldDataModule(crossval_count=1, seed=1)

    def get_callbacks(self) -> List[Callback]:
        if self.save_checkpoint:
            return [
                ModelCheckpoint(
                    dirpath=self.checkpoint_folder,
                    monitor="val_loss",
                    filename="checkpoint",
                    auto_insert_metric_name=False,
                    mode="min",
                ),
                *super().get_callbacks(),
            ]
        else:
            return super().get_callbacks()

    def get_additional_aml_run_tags(self) -> Dict[str, str]:
        return {"max_epochs": str(self.max_epochs)}



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from typing import Any, Dict, List, Optional, Tuple
from pathlib import Path

import param
import logging
from azureml.core import ScriptRunConfig
from azureml.train.hyperdrive import HyperDriveConfig
from pytorch_lightning import Callback, LightningDataModule, LightningModule
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler

from health_azure.utils import create_from_matching_params
from health_ml.deep_learning_config import DatasetParams, OptimizerParams, OutputParams, TrainerParams, WorkflowParams
from health_ml.experiment_config import ExperimentConfig, RunnerMode
from health_ml.utils.checkpoint_utils import get_best_checkpoint_path
from health_ml.utils.lr_scheduler import SchedulerWithWarmUp
from health_ml.utils.model_util import create_optimizer


class LightningContainer(WorkflowParams, DatasetParams, OutputParams, TrainerParams, OptimizerParams):
    """
    A LightningContainer contains all information to train a user-specified PyTorch Lightning model. The model that
    should be trained is returned by the `get_model` method. The training data must be returned in the form of
    a LightningDataModule, by the `get_data_module` method.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._model: Optional[LightningModule] = None
        self._model_name = type(self).__name__
        self.trained_weights_path: Optional[Path] = None
        # Number of nodes and the runner mode are read from the ExperimentConfig, and will be copied here
        self.num_nodes = 1
        self.runner_mode = RunnerMode.TRAIN

    def validate(self) -> None:
        WorkflowParams.validate(self)
        OptimizerParams.validate(self)
        TrainerParams.validate(self)

    def setup(self) -> None:
        """
        This method is called as one of the first operations of the training/testing workflow, before any other
        operations on the present object. At the point when called, the datasets are already available in
        the locations given by self.local_datasets. Use this method to prepare datasets or data loaders, for example.
        """
        pass

    def create_model(self) -> LightningModule:  # type: ignore
        """
        This method must create the actual Lightning model that will be trained. It can read out parameters from the
        container and pass them into the model, for example.
        """
        pass

    def get_data_module(self) -> LightningDataModule:
        """
        Gets the data that is used for the training, validation, and test steps.
        This should read datasets from the self.local_datasets folder or download from a web location.
        The format of the data is not specified any further.

        :return: A LightningDataModule
        """
        return None  # type: ignore

    def get_eval_data_module(self) -> LightningDataModule:
        """
        Gets the data that is used when evaluating the model on a new dataset.
        This data module should read datasets from the self.local_datasets folder or download from a web location.
        Only the test dataloader is used, hence the method needs to put all data into the test dataloader, rather
        than splitting into train/val/test.

        :return: A LightningDataModule
        """
        return None  # type: ignore

    def get_trainer_arguments(self) -> Dict[str, Any]:
        """
        Gets additional parameters that will be passed on to the PyTorch Lightning trainer.
        """
        return dict()

    def get_callbacks(self) -> List[Callback]:
        """
        Gets additional callbacks that the trainer should use when training this model.
        """
        return []

    def get_parameter_tuning_config(self, run_config: ScriptRunConfig) -> HyperDriveConfig:  # type: ignore
        """
        Returns a configuration for hyperparameter tuning via AzureML's Hyperdrive capability.
        Hyperparameter tuning can be triggered on the commandline via the "--hyperdrive" flag.
        Override this method in your LightningContainer to use hyperparameter tuning.

        The HyperDriveConfig config object needs to specify which parameters should be searched over, and which
        metric should be monitored.

        :param run_config: The ScriptRunConfig object that needs to be passed into the constructor of
            HyperDriveConfig.
        """
        raise NotImplementedError(
            "Parameter search is not implemented. Please override 'get_parameter_tuning_config' "
            "in your model container."
        )

    def get_parameter_tuning_args(self) -> Dict[str, Any]:
        """
        Returns a dictionary of hyperperameter argument names and values as expected by a AML SDK v2 job
        to perform hyperparameter search
        """
        raise NotImplementedError(
            "Parameter search is not implemented. Please override 'get_parameter_tuning_args' "
            "in your model container."
        )

    def update_experiment_config(self, experiment_config: ExperimentConfig) -> None:
        """
        This method allows overriding ExperimentConfig parameters from within a LightningContainer.
        It is called right after the ExperimentConfig and container are initialised.
        Be careful when using class parameters to override these values. If the parameter names clash,
        CLI values will be consumed by the ExperimentConfig, but container parameters will keep their defaults.
        This can be avoided by always using unique parameter names.
        Also note that saving a reference to `experiment_config` and updating its attributes at any other
        point may lead to unexpected behaviour.

        :param experiment_config: The initialised ExperimentConfig whose parameters to override in-place.
        """
        pass

    def before_training_on_global_rank_zero(self) -> None:
        """
        A hook that will be called before starting model training, before creating the Lightning Trainer object.
        In distributed training, this is only run on global rank zero (i.e, on the process that runs on node 0, GPU 0).
        The order in which hooks are called is: before_training_on_global_rank_zero, before_training_on_local_rank_zero,
        before_training_on_all_ranks.
        """
        pass

    def before_training_on_local_rank_zero(self) -> None:
        """
        A hook that will be called before starting model training.
        In distributed training, this hook will be called once per node (i.e., whenever the LOCAL_RANK environment
        variable is zero).
        The order in which hooks are called is: before_training_on_global_rank_zero, before_training_on_local_rank_zero,
        before_training_on_all_ranks.
        """
        pass

    def before_training_on_all_ranks(self) -> None:
        """
        A hook that will be called before starting model training.
        In distributed training, this hook will be called on all ranks (i.e., once per GPU).
        The order in which hooks are called is: before_training_on_global_rank_zero, before_training_on_local_rank_zero,
        before_training_on_all_ranks.
        """
        pass

    def get_checkpoint_to_test(self) -> Path:
        """Returns the path of the model checkpoint that should be used for testing/inference. By default, this will
        return the checkpoint that is written in the last training epoch. Override this method if you implement a
        custom checkpointing logic, for example if you added a model checkpoint callback that looks at validation
        accuracy.

        :return: The path of the checkpoint file that should be used for inference.
        """
        return get_best_checkpoint_path(self.checkpoint_folder)

    # The code from here on does not need to be modified.

    @property
    def model(self) -> LightningModule:
        """
        Returns the PyTorch Lightning module that the present container object manages.

        :return: A PyTorch Lightning module
        """
        if self._model is None:
            raise ValueError("No Lightning module has been set yet.")
        return self._model

    def create_lightning_module_and_store(self) -> None:
        """
        Creates the Lightning model
        """
        self._model = self.create_model()
        if isinstance(self._model, LightningModuleWithOptimizer):
            self._model._optimizer_params = create_from_matching_params(self, OptimizerParams)
            self._model._trainer_params = create_from_matching_params(self, TrainerParams)

    def get_hyperdrive_config(self) -> Optional[HyperDriveConfig]:
        """
        Returns the HyperDrive config for either hyperparameter tuning, cross validation, or running with
        different seeds.

        :return: A configuration object for HyperDrive
        """
        if self.hyperdrive:
            return self.get_parameter_tuning_config(ScriptRunConfig(source_directory=""))
        if self.is_crossvalidation_parent_run:
            return self.get_crossval_hyperdrive_config()
        if self.different_seeds > 0:
            return self.get_different_seeds_hyperdrive_config()
        return None

    def get_hyperparam_args(self) -> Optional[Dict[str, Any]]:
        """
        Returns a dictionary of hyperparameter search arguments that will be passed to an AML v2 command to
        enable either hyperparameter tuning,  cross validation, or running with different seeds.

        :return: A dictionary of hyperparameter search arguments and values.
        """
        if self.hyperdrive:
            return self.get_parameter_tuning_args()
        if self.is_crossvalidation_parent_run:
            return self.get_crossval_hyperparam_args_v2()
        if self.different_seeds > 0:
            return self.get_grid_hyperparam_args_v2()
        return None

    def load_model_checkpoint(self, checkpoint_path: Path) -> None:
        """
        Load a checkpoint from the given path. We need to define a separate method since pytorch lightning cannot
        access the _model attribute to modify it.
        """
        if self._model is None:
            raise ValueError("No Lightning module has been set yet.")
        self._model = type(self._model).load_from_checkpoint(checkpoint_path=str(checkpoint_path))

    def __str__(self) -> str:
        """Returns a string describing the present object, as a list of key: value strings."""
        arguments_str = "\nContainer:\n"
        # Avoid callable params, the bindings that are printed out can be humongous.
        # Avoid dataframes
        skip_params = {
            name for name, value in self.param.params().items() if isinstance(value, (param.Callable, param.DataFrame))
        }
        for key, value in self.param.get_param_values():
            if key not in skip_params:
                arguments_str += f"\t{key:40}: {value}\n"
        # Print out all other separate vars that are not under the guidance of the params library,
        # skipping the two that are introduced by params
        skip_vars = {"param", "initialized"}
        for key, value in vars(self).items():
            if key not in skip_vars and key[0] != "_":
                arguments_str += f"\t{key:40}: {value}\n"
        return arguments_str

    def has_custom_test_step(self) -> bool:
        """
        Determines if the lightning module has a custom test step so that the runner can determine whether to
        run inference or skip it.
        """
        return type(self.model).test_step != LightningModule.test_step

    @property
    def effective_experiment_name(self) -> str:
        """Returns the name of the AzureML experiment that should be used. This is taken from the commandline
        argument `experiment`, falling back to the model class name if not set."""
        return self.experiment or self.model_name

    def get_additional_aml_run_tags(self) -> Dict[str, str]:
        """Returns a dictionary of tags that should be added to the AzureML run."""
        return {}

    def get_additional_environment_variables(self) -> Dict[str, str]:
        """Returns a dictionary of environment variables that should be added to the AzureML run."""
        return {}

    def on_run_extra_validation_epoch(self) -> None:
        if hasattr(self.model, "on_run_extra_validation_epoch"):
            assert self._model, "Model is not initialized."
            self.model.on_run_extra_validation_epoch()  # type: ignore
        else:
            logging.warning(
                "Hook `on_run_extra_validation_epoch` is not implemented by lightning module."
                "The extra validation epoch won't produce any extra outputs."
            )

    def set_model_variant(self, variant_name: str) -> None:
        """Choose which variant of the model to use. A variant can for example have a different number of layers
        compared to the base model. This method is called by the runner to set the variant that should be used, passing
        in the variant name. A typical implement would set parameters of the object, based on the value of the
        `variant_name` argument.

        :param variant_name: The name of the model variant that should be set.
        """
        pass


class LightningModuleWithOptimizer(LightningModule):
    """
    A base class that supplies a method to configure optimizers and LR schedulers. To use this in your model,
    inherit from this class instead of from LightningModule.
    If this class is used, all configuration options for the optimizers and LR schedulers will be also available as
    commandline arguments (for example, you can supply the hi-ml runner with "--l_rate=1e-2" to change the learning
    rate.
    """

    # These fields will be set by the LightningContainer when the model is created.
    _optimizer_params = OptimizerParams()
    _trainer_params = TrainerParams()

    def configure_optimizers(self) -> Tuple[List[Optimizer], List[_LRScheduler]]:
        """
        This is the default implementation of the method that provides the optimizer and LR scheduler for
        PyTorch Lightning. It reads out the optimizer and scheduler settings from the model fields,
        and creates the two objects.
        Override this method for full flexibility to define any optimizer and scheduler.
        :return: A tuple of (optimizer, LR scheduler)
        """
        optimizer = create_optimizer(self._optimizer_params, self.parameters())
        l_rate_scheduler = SchedulerWithWarmUp(
            self._optimizer_params, optimizer, num_epochs=self._trainer_params.max_epochs
        )
        return [optimizer], [l_rate_scheduler]
